from typing import Any, Self, TypedDict, Literal
from openai.types.chat.chat_completion_reasoning_effort import ChatCompletionReasoningEffort
from pydantic import BaseModel, Field, model_validator

from ..mime import MIMEData
from ..modalities import Modality


class EnhanceSchemaConfig(BaseModel):
    allow_field_description_update: bool = False  # Whether to allow the llm to update the description of existing fields
    allow_system_prompt_update: bool = True  # Whether to allow the llm to update the system prompt
    allow_reasoning_field_toggle: bool = False  # Whether to allow the llm to toggle the reasoning for fields

    # Model validator
    @model_validator(mode="after")
    def check_at_least_one_tool_allowed(self) -> Self:
        if not any([self.allow_field_description_update, self.allow_system_prompt_update, self.allow_reasoning_field_toggle]):
            raise ValueError("At least one tool must be allowed")
        return self


# Define a typed Dict for EnhanceSchemaConfig (for now it is kind static, but we will add more flexibility in the future)
class EnhanceSchemaConfigDict(TypedDict, total=False):
    allow_field_description_update: bool
    allow_system_prompt_update: bool
    allow_reasoning_field_toggle: bool


class EnhanceSchemaRequest(BaseModel):
    """
    The request body for enhancing a JSON Schema.
    """

    documents: list[MIMEData]
    ground_truths: list[dict[str, Any]] | None = None
    model: str = "gpt-4o-mini"
    temperature: float = 0.0
    reasoning_effort: ChatCompletionReasoningEffort = "medium"
    modality: Modality
    """The modality of the document to load."""

    image_resolution_dpi: int = 96
    browser_canvas: Literal['A3', 'A4', 'A5'] = 'A4'
    """The image operations to apply to the document."""

    stream: bool = False
    """Whether to stream the response."""

    tools_config: EnhanceSchemaConfig = Field(default_factory=EnhanceSchemaConfig, description="The configuration for the tools to use")

    json_schema: dict[str, Any]
    instructions: str | None = None
    flat_likelihoods: list[dict[str, float]] | dict[str, float] | None = None



from typing import Any, Dict, List, Optional, TypedDict, Union, Literal
from io import IOBase
from pathlib import Path

import PIL.Image
from pydantic import HttpUrl

from .._resource import AsyncAPIResource, SyncAPIResource
from ..types.standards import PreparedRequest
from ..types.evals import (
    Evaluation,
    EvaluationDocument,
    Iteration,
    DistancesResult,
    PredictionData,
    AddIterationFromJsonlRequest,
    DocumentItem,
    UpdateEvaluationDocumentRequest,
    PredictionMetadata,
    CreateIterationRequest,
)
from ..types.jobs.base import InferenceSettings

from ..types.mime import MIMEData
from .._utils.mime import prepare_mime_document
from ..types.modalities import Modality
from openai.types.chat.chat_completion_reasoning_effort import ChatCompletionReasoningEffort

from tqdm import tqdm


class DeleteResponse(TypedDict):
    """Response from a delete operation"""

    success: bool
    id: str


class ExportResponse(TypedDict):
    """Response from an export operation"""

    success: bool
    path: str


class EvalsMixin:
    def prepare_create(
        self,
        name: str,
        json_schema: Dict[str, Any],
        project_id: str | None = None,
        documents: List[EvaluationDocument] = [],
        iterations: List[Iteration] = [],
        default_inference_settings: Optional[InferenceSettings] = None,
    ) -> PreparedRequest:
        eval_data = Evaluation(
            name=name,
            json_schema=json_schema,
            project_id=project_id if project_id else "default_spreadsheets",
            documents=documents,
            iterations=iterations,
            default_inference_settings=default_inference_settings,
        )
        return PreparedRequest(method="POST", url="/v1/evals", data=eval_data.model_dump(exclude_none=True, mode="json"))

    def prepare_get(self, evaluation_id: str) -> PreparedRequest:
        return PreparedRequest(method="GET", url=f"/v1/evals/{evaluation_id}")

    def prepare_update(
        self,
        evaluation_id: str,
        name: Optional[str] = None,
        project_id: Optional[str] = None,
        json_schema: Optional[Dict[str, Any]] = None,
        documents: Optional[List[EvaluationDocument]] = None,
        iterations: Optional[List[Iteration]] = None,
        default_inference_settings: Optional[InferenceSettings] = None,
    ) -> PreparedRequest:
        """
        Prepare a request to update an evaluation with partial updates.

        Only the provided fields will be updated. Fields set to None will be excluded from the update.
        """
        # Build a dictionary with only the provided fields
        update_data = {}
        if name is not None:
            update_data["name"] = name
        if project_id is not None:
            update_data["project_id"] = project_id
        if json_schema is not None:
            update_data["json_schema"] = json_schema
        if documents is not None:
            update_data["documents"] = [doc.model_dump(exclude_none=True, mode="json") for doc in documents]
        if iterations is not None:
            update_data["iterations"] = [iter.model_dump(exclude_none=True, mode="json") for iter in iterations]
        if default_inference_settings is not None:
            update_data["default_inference_settings"] = default_inference_settings.model_dump(exclude_none=True, mode="json")

        return PreparedRequest(method="PATCH", url=f"/v1/evals/{evaluation_id}", data=update_data)

    def prepare_list(self, project_id: Optional[str] = None) -> PreparedRequest:
        params = {}
        if project_id:
            params["project_id"] = project_id
        return PreparedRequest(method="GET", url="/v1/evals", params=params)

    def prepare_delete(self, id: str) -> PreparedRequest:
        return PreparedRequest(method="DELETE", url=f"/v1/evals/{id}")


class DocumentsMixin:
    def prepare_get(self, evaluation_id: str, document_id: str) -> PreparedRequest:
        return PreparedRequest(method="GET", url=f"/v1/evals/{evaluation_id}/documents/{document_id}")

    def prepare_create(self, evaluation_id: str, document: MIMEData, annotation: Dict[str, Any]) -> PreparedRequest:
        # Serialize the MIMEData

        document_item = DocumentItem(mime_data=document, annotation=annotation, annotation_metadata=None)

        return PreparedRequest(method="POST", url=f"/v1/evals/{evaluation_id}/documents", data=document_item.model_dump(mode="json"))

    def prepare_list(self, evaluation_id: str, filename: Optional[str] = None) -> PreparedRequest:
        params = {}
        if filename:
            params["filename"] = filename
        return PreparedRequest(method="GET", url=f"/v1/evals/{evaluation_id}/documents", params=params)

    def prepare_update(self, evaluation_id: str, document_id: str, annotation: Dict[str, Any]) -> PreparedRequest:
        update_request = UpdateEvaluationDocumentRequest(annotation=annotation, annotation_metadata=None)

        return PreparedRequest(method="PUT", url=f"/v1/evals/{evaluation_id}/documents/{document_id}", data=update_request.model_dump(mode="json", exclude_none=True))

    def prepare_delete(self, evaluation_id: str, document_id: str) -> PreparedRequest:
        return PreparedRequest(method="DELETE", url=f"/v1/evals/{evaluation_id}/documents/{document_id}")


class IterationsMixin:
    def prepare_get(self, iteration_id: str) -> PreparedRequest:
        return PreparedRequest(method="GET", url=f"/v1/evals/iterations/{iteration_id}")

    def prepare_list(self, evaluation_id: str, model: Optional[str] = None) -> PreparedRequest:
        params = {}
        if model:
            params["model"] = model
        return PreparedRequest(method="GET", url=f"/v1/evals/{evaluation_id}/iterations", params=params)

    def prepare_create(
        self,
        evaluation_id: str,
        model: str,
        json_schema: Optional[Dict[str, Any]] = None,
        temperature: float = 0.0,
        modality: Modality = "native",
        reasoning_effort: ChatCompletionReasoningEffort = "medium",
        image_resolution_dpi: int = 96,
        browser_canvas: Literal['A3', 'A4', 'A5'] = 'A4',
        n_consensus: int = 1,
    ) -> PreparedRequest:
        props = InferenceSettings(
            model=model,
            temperature=temperature,
            modality=modality,
            reasoning_effort=reasoning_effort,
            image_resolution_dpi=image_resolution_dpi,
            browser_canvas=browser_canvas,
            n_consensus=n_consensus,
        )

        perform_iteration_request = CreateIterationRequest(inference_settings=props, json_schema=json_schema)

        return PreparedRequest(method="POST", url=f"/v1/evals/{evaluation_id}/iterations/create", data=perform_iteration_request.model_dump(exclude_none=True, mode="json"))

    def prepare_update(
        self, iteration_id: str, json_schema: Dict[str, Any], model: str, temperature: float = 0.0, image_resolution_dpi: int = 96, browser_canvas: Literal['A3', 'A4', 'A5'] = 'A4'
    ) -> PreparedRequest:
        inference_settings = InferenceSettings(
            model=model,
            temperature=temperature,
            image_resolution_dpi=image_resolution_dpi,
            browser_canvas=browser_canvas,
        )

        iteration_data = Iteration(id=iteration_id, json_schema=json_schema, inference_settings=inference_settings, predictions=[])

        return PreparedRequest(method="PUT", url=f"/v1/evals/iterations/{iteration_id}", data=iteration_data.model_dump(exclude_none=True, mode="json"))

    def prepare_delete(self, iteration_id: str) -> PreparedRequest:
        return PreparedRequest(method="DELETE", url=f"/v1/evals/iterations/{iteration_id}")

    def prepare_compute_distances(self, iteration_id: str, document_id: str) -> PreparedRequest:
        return PreparedRequest(method="GET", url=f"/v1/evals/iterations/{iteration_id}/compute_distances/{document_id}")


class Evals(SyncAPIResource, EvalsMixin):
    """Evals API wrapper"""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.documents = Documents(self._client)
        self.iterations = Iterations(self._client)

    def create(self, name: str, json_schema: Dict[str, Any], project_id: str | None = None) -> Evaluation:
        """
        Create a new evaluation.

        Args:
            name: The name of the evaluation
            json_schema: The JSON schema for the evaluation
            project_id: The project ID to associate with the evaluation

        Returns:
            Evaluation: The created evaluation
        Raises:
            HTTPException if the request fails
        """
        request = self.prepare_create(name, json_schema, project_id)
        response = self._client._prepared_request(request)
        return Evaluation(**response)

    def get(self, evaluation_id: str) -> Evaluation:
        """
        Get an evaluation by ID.

        Args:
            evaluation_id: The ID of the evaluation to retrieve

        Returns:
            Evaluation: The evaluation
        Raises:
            HTTPException if the request fails
        """
        request = self.prepare_get(evaluation_id)
        response = self._client._prepared_request(request)
        return Evaluation(**response)

    def update(
        self,
        evaluation_id: str,
        name: Optional[str] = None,
        project_id: Optional[str] = None,
        json_schema: Optional[Dict[str, Any]] = None,
        documents: Optional[List[EvaluationDocument]] = None,
        iterations: Optional[List[Iteration]] = None,
        default_inference_settings: Optional[InferenceSettings] = None,
    ) -> Evaluation:
        """
        Update an evaluation with partial updates.

        Args:
            evaluation_id: The ID of the evaluation to update
            name: Optional new name for the evaluation
            project_id: Optional new project ID
            json_schema: Optional new JSON schema
            documents: Optional list of documents to update
            iterations: Optional list of iterations to update
            default_inference_settings: Optional annotation properties

        Returns:
            Evaluation: The updated evaluation
        Raises:
            HTTPException if the request fails
        """
        request = self.prepare_update(
            evaluation_id=evaluation_id, name=name, project_id=project_id, json_schema=json_schema, documents=documents, iterations=iterations, default_inference_settings=default_inference_settings
        )
        response = self._client._prepared_request(request)
        return Evaluation(**response)

    def list(self, project_id: str) -> List[Evaluation]:
        """
        List evaluations for a project.

        Args:
            project_id: The project ID to list evaluations for

        Returns:
            List[Evaluation]: List of evaluations
        Raises:
            HTTPException if the request fails
        """
        request = self.prepare_list(project_id)
        response = self._client._prepared_request(request)
        return [Evaluation(**item) for item in response.get("data", [])]

    def delete(self, evaluation_id: str) -> DeleteResponse:
        """
        Delete an evaluation.

        Args:
            evaluation_id: The ID of the evaluation to delete

        Returns:
            DeleteResponse: The response containing success status and ID
        Raises:
            HTTPException if the request fails
        """
        request = self.prepare_delete(evaluation_id)
        return self._client._prepared_request(request)


class Documents(SyncAPIResource, DocumentsMixin):
    """Documents API wrapper for evaluations"""

    def create(self, evaluation_id: str, document: Union[Path, str, IOBase, MIMEData, PIL.Image.Image, HttpUrl], annotation: Dict[str, Any]) -> EvaluationDocument:
        """
        Create a document for an evaluation.

        Args:
            evaluation_id: The ID of the evaluation
            document: The document to process. Can be:
                - A file path (Path or str)
                - A file-like object (IOBase)
                - A MIMEData object
                - A PIL Image object
                - A URL (HttpUrl)
            annotation: The ground truth for the document

        Returns:
            EvaluationDocument: The created document
        Raises:
            HTTPException if the request fails
        """
        # Convert document to MIME data format
        mime_document: MIMEData = prepare_mime_document(document)

        # Let prepare_create handle the serialization
        request = self.prepare_create(evaluation_id, mime_document, annotation)
        response = self._client._prepared_request(request)
        return EvaluationDocument(**response)

    def list(self, evaluation_id: str, filename: Optional[str] = None) -> List[EvaluationDocument]:
        """
        List documents for an evaluation.

        Args:
            evaluation_id: The ID of the evaluation
            filename: Optional filename to filter by

        Returns:
            List[EvaluationDocument]: List of documents
        Raises:
            HTTPException if the request fails
        """
        request = self.prepare_list(evaluation_id, filename)
        response = self._client._prepared_request(request)
        return [EvaluationDocument(**item) for item in response.get("data", [])]

    def get(self, evaluation_id: str, document_id: str) -> EvaluationDocument:
        """
        Get a document by ID.

        Args:
            evaluation_id: The ID of the evaluation
            document_id: The ID of the document

        Returns:
            EvaluationDocument: The document
        Raises:
            HTTPException if the request fails
        """
        request = self.prepare_get(evaluation_id, document_id)
        response = self._client._prepared_request(request)
        return EvaluationDocument(**response)

    def update(self, evaluation_id: str, document_id: str, annotation: Dict[str, Any]) -> EvaluationDocument:
        """
        Update a document.

        Args:
            evaluation_id: The ID of the evaluation
            document_id: The ID of the document
            annotation: The ground truth for the document

        Returns:
            EvaluationDocument: The updated document
        Raises:
            HTTPException if the request fails
        """
        request = self.prepare_update(evaluation_id, document_id, annotation)
        response = self._client._prepared_request(request)
        return EvaluationDocument(**response)

    def delete(self, evaluation_id: str, document_id: str) -> DeleteResponse:
        """
        Delete a document.

        Args:
            evaluation_id: The ID of the evaluation
            document_id: The ID of the document

        Returns:
            DeleteResponse: The response containing success status and ID
        Raises:
            HTTPException if the request fails
        """
        request = self.prepare_delete(evaluation_id, document_id)
        return self._client._prepared_request(request)


class Iterations(SyncAPIResource, IterationsMixin):
    """Iterations API wrapper for evaluations"""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def list(self, evaluation_id: str, model: Optional[str] = None) -> List[Iteration]:
        """
        List iterations for an evaluation.

        Args:
            evaluation_id: The ID of the evaluation
            model: Optional model to filter by

        Returns:
            List[Iteration]: List of iterations
        Raises:
            HTTPException if the request fails
        """
        request = self.prepare_list(evaluation_id, model)
        response = self._client._prepared_request(request)
        return [Iteration(**item) for item in response.get("data", [])]

    def create(
        self,
        evaluation_id: str,
        model: str,
        temperature: float = 0.0,
        modality: Modality = "native",
        json_schema: Optional[Dict[str, Any]] = None,
        reasoning_effort: ChatCompletionReasoningEffort = "medium",
        image_resolution_dpi: int = 96,
        browser_canvas: Literal['A3', 'A4', 'A5'] = 'A4',
        n_consensus: int = 1,
    ) -> Iteration:
        """
        Create a new iteration for an evaluation.

        Args:
            evaluation_id: The ID of the evaluation
            json_schema: The JSON schema for the iteration (if not set, we use the one of the eval)
            model: The model to use for the iteration
            temperature: The temperature to use for the model
            modality: The modality to use (text, image, etc.)
            reasoning_effort: The reasoning effort setting for the model (auto, low, medium, high)
            image_resolution_dpi: The DPI of the image. Defaults to 96.
            browser_canvas: The canvas size of the browser. Must be one of:
                - "A3" (11.7in x 16.54in)
                - "A4" (8.27in x 11.7in) 
                - "A5" (5.83in x 8.27in)
                Defaults to "A4".
            n_consensus: Number of consensus iterations to perform

        Returns:
            Iteration: The created iteration
        Raises:
            HTTPException if the request fails
        """
        request = self.prepare_create(
            evaluation_id=evaluation_id,
            json_schema=json_schema,
            model=model,
            temperature=temperature,
            modality=modality,
            reasoning_effort=reasoning_effort,
            image_resolution_dpi=image_resolution_dpi,
            browser_canvas=browser_canvas,
            n_consensus=n_consensus,
        )
        response = self._client._prepared_request(request)
        return Iteration(**response)

    def delete(self, iteration_id: str) -> DeleteResponse:
        """
        Delete an iteration.

        Args:
            iteration_id: The ID of the iteration

        Returns:
            DeleteResponse: The response containing success status and ID
        Raises:
            HTTPException if the request fails
        """
        request = self.prepare_delete(iteration_id)
        return self._client._prepared_request(request)

    def compute_distances(self, iteration_id: str, document_id: str) -> DistancesResult:
        """
        Get distances for a document in an iteration.

        Args:
            iteration_id: The ID of the iteration
            document_id: The ID of the document

        Returns:
            DistancesResult: The distances
        Raises:
            HTTPException if the request fails
        """
        request = self.prepare_compute_distances(iteration_id, document_id)
        response = self._client._prepared_request(request)
        return DistancesResult(**response)


class AsyncEvals(AsyncAPIResource, EvalsMixin):
    """Async Evals API wrapper"""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.documents = AsyncDocuments(self._client)
        self.iterations = AsyncIterations(self._client)

    async def create(self, name: str, json_schema: Dict[str, Any], project_id: str | None = None) -> Evaluation:
        """
        Create a new evaluation.

        Args:
            name: The name of the evaluation
            json_schema: The JSON schema for the evaluation
            project_id: The project ID to associate with the evaluation

        Returns:
            Evaluation: The created evaluation
        Raises:
            HTTPException if the request fails
        """
        request = self.prepare_create(name, json_schema, project_id)
        response = await self._client._prepared_request(request)
        return Evaluation(**response)

    async def get(self, evaluation_id: str) -> Evaluation:
        """
        Get an evaluation by ID.

        Args:
            evaluation_id: The ID of the evaluation to retrieve

        Returns:
            Evaluation: The evaluation
        Raises:
            HTTPException if the request fails
        """
        request = self.prepare_get(evaluation_id)
        response = await self._client._prepared_request(request)
        return Evaluation(**response)

    async def update(
        self,
        evaluation_id: str,
        name: Optional[str] = None,
        project_id: Optional[str] = None,
        json_schema: Optional[Dict[str, Any]] = None,
        documents: Optional[List[EvaluationDocument]] = None,
        iterations: Optional[List[Iteration]] = None,
        default_inference_settings: Optional[InferenceSettings] = None,
    ) -> Evaluation:
        """
        Update an evaluation with partial updates.

        Args:
            id: The ID of the evaluation to update
            name: Optional new name for the evaluation
            project_id: Optional new project ID
            json_schema: Optional new JSON schema
            documents: Optional list of documents to update
            iterations: Optional list of iterations to update
            default_inference_settings: Optional annotation properties

        Returns:
            Evaluation: The updated evaluation
        Raises:
            HTTPException if the request fails
        """
        request = self.prepare_update(
            evaluation_id=evaluation_id, name=name, project_id=project_id, json_schema=json_schema, documents=documents, iterations=iterations, default_inference_settings=default_inference_settings
        )
        response = await self._client._prepared_request(request)
        return Evaluation(**response)

    async def list(self, project_id: Optional[str] = None) -> List[Evaluation]:
        """
        List evaluations for a project.

        Args:
            project_id: The project ID to list evaluations for

        Returns:
            List[Evaluation]: List of evaluations
        Raises:
            HTTPException if the request fails
        """
        request = self.prepare_list(project_id)
        response = await self._client._prepared_request(request)
        return [Evaluation(**item) for item in response.get("data", [])]

    async def delete(self, evaluation_id: str) -> DeleteResponse:
        """
        Delete an evaluation.

        Args:
            evaluation_id: The ID of the evaluation to delete

        Returns:
            DeleteResponse: The response containing success status and ID
        Raises:
            HTTPException if the request fails
        """
        request = self.prepare_delete(evaluation_id)
        return await self._client._prepared_request(request)


class AsyncDocuments(AsyncAPIResource, DocumentsMixin):
    """Async Documents API wrapper for evaluations"""

    async def create(self, evaluation_id: str, document: Union[Path, str, IOBase, MIMEData, PIL.Image.Image, HttpUrl], annotation: Dict[str, Any]) -> EvaluationDocument:
        """
        Create a document for an evaluation.

        Args:
            evaluation_id: The ID of the evaluation
            document: The document to process. Can be:
                - A file path (Path or str)
                - A file-like object (IOBase)
                - A MIMEData object
                - A PIL Image object
                - A URL (HttpUrl)
            annotation: The ground truth for the document

        Returns:
            EvaluationDocument: The created document
        Raises:
            HTTPException if the request fails
        """
        # Convert document to MIME data format
        mime_document: MIMEData = prepare_mime_document(document)

        # Let prepare_create handle the serialization
        request = self.prepare_create(evaluation_id, mime_document, annotation)
        response = await self._client._prepared_request(request)
        return EvaluationDocument(**response)

    async def list(self, evaluation_id: str, filename: Optional[str] = None) -> List[EvaluationDocument]:
        """
        List documents for an evaluation.

        Args:
            evaluation_id: The ID of the evaluation
            filename: Optional filename to filter by

        Returns:
            List[EvaluationDocument]: List of documents
        Raises:
            HTTPException if the request fails
        """
        request = self.prepare_list(evaluation_id, filename)
        response = await self._client._prepared_request(request)
        return [EvaluationDocument(**item) for item in response.get("data", [])]

    async def update(self, evaluation_id: str, document_id: str, annotation: Dict[str, Any]) -> EvaluationDocument:
        """
        Update a document.

        Args:
            evaluation_id: The ID of the evaluation
            document_id: The ID of the document
            annotation: The ground truth for the document

        Returns:
            EvaluationDocument: The updated document
        Raises:
            HTTPException if the request fails
        """
        request = self.prepare_update(evaluation_id, document_id, annotation)
        response = await self._client._prepared_request(request)
        return EvaluationDocument(**response)

    async def delete(self, evaluation_id: str, document_id: str) -> DeleteResponse:
        """
        Delete a document.

        Args:
            evaluation_id: The ID of the evaluation
            document_id: The ID of the document

        Returns:
            DeleteResponse: The response containing success status and ID
        Raises:
            HTTPException if the request fails
        """
        request = self.prepare_delete(evaluation_id, document_id)
        return await self._client._prepared_request(request)


class AsyncIterations(AsyncAPIResource, IterationsMixin):
    """Async Iterations API wrapper for evaluations"""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    async def get(self, iteration_id: str) -> Iteration:
        """
        Get an iteration by ID.

        Args:
            iteration_id: The ID of the iteration

        Returns:
            Iteration: The iteration
        Raises:
            HTTPException if the request fails
        """
        request = self.prepare_get(iteration_id)
        response = await self._client._prepared_request(request)
        return Iteration(**response)

    async def list(self, evaluation_id: str, model: Optional[str] = None) -> List[Iteration]:
        """
        List iterations for an evaluation.

        Args:
            evaluation_id: The ID of the evaluation
            model: Optional model to filter by

        Returns:
            List[Iteration]: List of iterations
        Raises:
            HTTPException if the request fails
        """
        request = self.prepare_list(evaluation_id, model)
        response = await self._client._prepared_request(request)
        return [Iteration(**item) for item in response.get("data", [])]

    async def create(
        self,
        evaluation_id: str,
        model: str,
        temperature: float = 0.0,
        modality: Modality = "native",
        json_schema: Optional[Dict[str, Any]] = None,
        reasoning_effort: ChatCompletionReasoningEffort = "medium",
        image_resolution_dpi: int = 96,
        browser_canvas: Literal['A3', 'A4', 'A5'] = 'A4',
        n_consensus: int = 1,
    ) -> Iteration:
        """
        Create a new iteration for an evaluation.

        Args:
            evaluation_id: The ID of the evaluation
            json_schema: The JSON schema for the iteration
            model: The model to use for the iteration
            temperature: The temperature to use for the model
            modality: The modality to use (text, image, etc.)
            reasoning_effort: The reasoning effort setting for the model (auto, low, medium, high)
            image_resolution_dpi: The DPI of the image. Defaults to 96.
            browser_canvas: The canvas size of the browser. Must be one of:
                - "A3" (11.7in x 16.54in)
                - "A4" (8.27in x 11.7in) 
                - "A5" (5.83in x 8.27in)
                Defaults to "A4".
            n_consensus: Number of consensus iterations to perform

        Returns:
            Iteration: The created iteration
        Raises:
            HTTPException if the request fails
        """
        request = self.prepare_create(
            evaluation_id=evaluation_id,
            json_schema=json_schema,
            model=model,
            temperature=temperature,
            modality=modality,
            reasoning_effort=reasoning_effort,
            image_resolution_dpi=image_resolution_dpi,
            browser_canvas=browser_canvas,
            n_consensus=n_consensus,
        )
        response = await self._client._prepared_request(request)
        return Iteration(**response)

    async def delete(self, iteration_id: str) -> DeleteResponse:
        """
        Delete an iteration.

        Args:
            iteration_id: The ID of the iteration

        Returns:
            DeleteResponse: The response containing success status and ID
        Raises:
            HTTPException if the request fails
        """
        request = self.prepare_delete(iteration_id)
        return await self._client._prepared_request(request)

    async def compute_distances(self, iteration_id: str, document_id: str) -> DistancesResult:
        """
        Get distances for a document in an iteration.

        Args:
            iteration_id: The ID of the iteration
            document_id: The ID of the document

        Returns:
            DistancesResult: The distances
        Raises:
            HTTPException if the request fails
        """
        request = self.prepare_compute_distances(iteration_id, document_id)
        response = await self._client._prepared_request(request)
        return DistancesResult(**response)



from typing import Literal, Optional, Self

from openai.types.chat.chat_completion_reasoning_effort import ChatCompletionReasoningEffort
from pydantic import BaseModel, Field, model_validator

from ..modalities import Modality

SelectionMode = Literal["all", "manual"]


# This is the input data for the prepare_dataset job
class PrepareDatasetInputData(BaseModel):
    dataset_id: Optional[str] = None
    schema_id: Optional[str] = None
    schema_data_id: Optional[str] = None

    selection_model: SelectionMode = "all"

    @model_validator(mode="after")
    def validate_input(self) -> Self:
        # The preference is:
        # 1. dataset_id
        # 2. schema_id
        # 3. schema_data_id
        if self.dataset_id is None and self.schema_id is None and self.schema_data_id is None:
            raise ValueError("At least one of dataset_id, schema_id, or schema_data_id must be provided")

        return self


# This is the input data for the split_dataset job
class DatasetSplitInputData(BaseModel):
    dataset_id: str
    train_size: Optional[int | float] = None
    eval_size: Optional[int | float] = None

    @model_validator(mode="after")
    def validate_input(self) -> Self:
        if self.train_size is not None and self.eval_size is not None:
            raise ValueError("train_size and eval_size cannot both be provided")
        return self


# This is the input data for the batch annotation job
class InferenceSettings(BaseModel):
    model: str = "gpt-4o-mini"
    temperature: float = 0.0
    modality: Modality = "native"
    reasoning_effort: ChatCompletionReasoningEffort = "medium"
    image_resolution_dpi: int = 96
    browser_canvas: Literal['A3', 'A4', 'A5'] = 'A4'
    n_consensus: int = Field(default=1, description="Number of consensus rounds to perform")


class AnnotationInputData(BaseModel):
    data_file: str
    schema_id: str
    inference_settings: InferenceSettings


# This is the input data for the evaluation job
class EvaluationInputData(BaseModel):
    eval_data_file: str
    schema_id: str
    inference_settings_1: InferenceSettings | None = None
    inference_settings_2: InferenceSettings


# from pydantic import BaseModel, Field, model_validator
# from typing import Literal, Optional, Any
# import datetime


# JobType = Literal["prompt-optimization", "annotate-files", "finetune-dataset", "webcrawl"]
# JobStatus = Literal["pending", "running", "completed", "failed"]
#### JOBS ####

# class JobTemplateCreateRequest(BaseModel):
#     job_type: JobType
#     default_input_data: dict = Field(default_factory=dict)
#     description: Optional[str] = None
#     cron: Optional[str] = None


# class JobTemplateDocument(BaseModel):
#     object: Literal["job_template"] = "job_template"
#     id: str
#     type: JobType
#     identity: Any | None = None
#     description: Optional[str] = None
#     default_input_data: dict = Field(default_factory=dict)
#     # For scheduled jobs, include a valid CRON expression (None for on-demand only jobs)
#     cron: Optional[str] = None
#     next_run: Optional[datetime.datetime] = None
#     created_at: Optional[datetime.datetime] = None
#     updated_at: Optional[datetime.datetime] = None
#     is_active: bool = True  # Change to status.

# class JobTemplateUpdateRequest(BaseModel):
#     cron: Optional[str] = None
#     default_input_data: Optional[dict] = None
#     description: Optional[str] = None
#     is_active: Optional[bool] = None    # Change to status.


#### EXECUTIONS ####

# class JobExecutionCreateRequest(BaseModel):
#     type: JobType
#     template_id: Optional[str] = None
#     input_data: dict = Field(default_factory=dict)

#     @model_validator(mode='before')
#     @classmethod
#     def validate_job_identifiers(cls, data: Any) -> Any:
#         if isinstance(data, dict):
#             if bool(data.get('job_type')) == bool(data.get('job_template_id')):
#                 raise ValueError("Either job_type or job_template_id must be provided")
#         return data

# class JobExecutionResponse(BaseModel):
#     id: str
#     template_id: Optional[str] = None
#     type: JobType
#     status: JobStatus
#     result: Optional[dict] = None
#     error: Optional[str] = None
#     created_at: Optional[datetime.datetime] = None
#     updated_at: Optional[datetime.datetime] = None

# class JobExecutionDocument(BaseModel):
#     object: Literal["job_execution"] = "job_execution"
#     id: str
#     template_id: Optional[str] = None
#     type: JobType
#     identity: Any | None = None
#     status: JobStatus
#     input_data_gcs_path: str
#     result: Optional[dict] = None
#     error: Optional[str] = None
#     created_at: Optional[datetime.datetime] = None
#     updated_at: Optional[datetime.datetime] = None
#     checkpoint: Any = None  # Useful for jobs that need to be resumed
#     checkpoint_data: Optional[dict] = None
#     needs: list[str] = Field(default_factory=list, description="list of jobs execution id that must be completed before this job can run")


# class Workflow(BaseModel):
#     name: str
#     jobs: list[JobExecutionDocument]



import datetime
import json

from jiter import from_json
from openai.types.chat.chat_completion_content_part_image_param import ChatCompletionContentPartImageParam, ImageURL
from openai.types.chat.chat_completion_content_part_param import ChatCompletionContentPartParam
from openai.types.chat.chat_completion_content_part_text_param import ChatCompletionContentPartTextParam
from openai.types.chat.parsed_chat_completion import ParsedChatCompletionMessage
from openai.types.completion_usage import CompletionTokensDetails, CompletionUsage, PromptTokensDetails
from openai.types.responses.easy_input_message_param import EasyInputMessageParam

# from openai.types.responses.response_input_param import ResponseInputParam
from openai.types.responses.response import Response
from openai.types.responses.response_input_image_param import ResponseInputImageParam
from openai.types.responses.response_input_message_content_list_param import ResponseInputMessageContentListParam
from openai.types.responses.response_input_param import ResponseInputItemParam
from openai.types.responses.response_input_text_param import ResponseInputTextParam

from ..types.chat import ChatCompletionUiformMessage
from ..types.documents.extractions import UiParsedChatCompletion, UiParsedChoice


def convert_to_openai_format(messages: list[ChatCompletionUiformMessage]) -> list[ResponseInputItemParam]:
    """
    Converts a list of ChatCompletionUiformMessage to the OpenAI ResponseInputParam format.

    Args:
        messages: List of chat messages in UIForm format

    Returns:
        Messages in OpenAI ResponseInputParam format for the Responses API
    """
    formatted_messages: list[ResponseInputItemParam] = []

    for message in messages:
        role = message["role"]
        content = message["content"]

        # Handle different content formats
        formatted_content: ResponseInputMessageContentListParam = []

        if isinstance(content, str):
            # Simple text content - provide direct string value
            formatted_content.append(ResponseInputTextParam(text=content, type="input_text"))
        elif isinstance(content, list):
            # Content is a list of parts
            for part in content:
                if part["type"] == "text":
                    formatted_content.append(ResponseInputTextParam(text=part["text"], type="input_text"))
                elif part["type"] == "image_url":
                    if "detail" in part["image_url"]:
                        detail = part["image_url"]["detail"]
                    else:
                        detail = "high"
                    formatted_content.append(ResponseInputImageParam(image_url=part["image_url"]["url"], type="input_image", detail=detail))
                else:
                    print(f"Not supported content type: {part['type']}... Skipping...")

        # Create Message structure which is one of the types in ResponseInputItemParam
        formatted_message = EasyInputMessageParam(role=role, content=formatted_content, type="message")

        formatted_messages.append(formatted_message)

    return formatted_messages


def convert_from_openai_format(messages: list[ResponseInputItemParam]) -> list[ChatCompletionUiformMessage]:
    """
    Converts messages from OpenAI ResponseInputParam format to ChatCompletionUiformMessage format.

    Args:
        messages: Messages in OpenAI ResponseInputParam format

    Returns:
        List of chat messages in UIForm format
    """
    formatted_messages: list[ChatCompletionUiformMessage] = []

    for message in messages:
        if "type" not in message:
            # The type is required by all other sub-types of ResponseInputItemParam except for EasyInputMessageParam and Message, which are messages.
            message["type"] = "message"

        if message["type"] != "message":
            print(f"Not supported message type: {message['type']}... Skipping...")
            continue
        role = message["role"]
        content = message["content"]
        formatted_content: str | list[ChatCompletionContentPartParam]

        if isinstance(content, str):
            formatted_content = content
        else:
            # Handle different content formats
            formatted_content = []
            for part in content:
                if part["type"] == "input_text":
                    formatted_content.append(ChatCompletionContentPartTextParam(text=part["text"], type="text"))
                elif part["type"] == "input_image":
                    image_url = part.get("image_url") or ""
                    image_detail = part.get("detail") or "high"
                    formatted_content.append(ChatCompletionContentPartImageParam(image_url=ImageURL(url=image_url, detail=image_detail), type="image_url"))
                else:
                    print(f"Not supported content type: {part['type']}... Skipping...")

        # Create message in UIForm format
        formatted_message = ChatCompletionUiformMessage(role=role, content=formatted_content)
        formatted_messages.append(formatted_message)

    return formatted_messages


def parse_openai_responses_response(response: Response) -> UiParsedChatCompletion:
    """
    Convert an OpenAI Response (Responses API) to UiParsedChatCompletion type.

    Args:
        response: Response from OpenAI Responses API

    Returns:
        Parsed response in UiParsedChatCompletion format
    """
    # Create the UiParsedChatCompletion object
    if response.usage:
        usage = CompletionUsage(
            prompt_tokens=response.usage.input_tokens,
            completion_tokens=response.usage.output_tokens,
            total_tokens=response.usage.total_tokens,
            prompt_tokens_details=PromptTokensDetails(
                cached_tokens=response.usage.input_tokens_details.cached_tokens,
            ),
            completion_tokens_details=CompletionTokensDetails(
                reasoning_tokens=response.usage.output_tokens_details.reasoning_tokens,
            ),
        )
    else:
        usage = None

    # Parse the ParsedChoice
    choices = []
    output_text = response.output_text
    result_object = from_json(bytes(output_text, "utf-8"), partial_mode=True)  # Attempt to parse the result even if EOF is reached

    choices.append(
        UiParsedChoice(
            index=0,
            message=ParsedChatCompletionMessage(
                role="assistant",
                content=json.dumps(result_object),
            ),
            finish_reason="stop",
        )
    )

    return UiParsedChatCompletion(
        id=response.id,
        choices=choices,
        created=int(datetime.datetime.now().timestamp()),
        model=response.model,
        object="chat.completion",
        likelihoods={},
        usage=usage,
    )



import datetime
from typing import Any, Literal, Optional

import nanoid  # type: ignore
from openai.types.chat import ChatCompletion
from openai.types.chat.chat_completion_reasoning_effort import ChatCompletionReasoningEffort
from pydantic import BaseModel, Field, computed_field

from uiform.types.chat import ChatCompletionUiformMessage
from uiform.types.documents.extractions import UiParsedChatCompletion

from .._utils.usage.usage import compute_cost_from_model, compute_cost_from_model_with_breakdown, CostBreakdown
from .ai_models import Amount
from .modalities import Modality

ValidationsState = Literal["pending", "validated", "invalid"]


class ExtractionSource(BaseModel):
    type: Literal["api", "annotation","processor", "automation.link", "automation.mailbox", "automation.cron", "automation.outlook", "automation.endpoint", "schema.extract"] = Field(
        description="Type of extraction"
    )
    id: str | None = Field(default=None, description="ID the trigger of the extraction")


ExtractionSteps = str | Literal['initialization', 'prepare_messages', 'yield_first_token', 'completion']  # Steps are meant to not overlap


class ExtractionTimingStep(BaseModel):
    name: ExtractionSteps
    duration: float  # in seconds
    notes: str | None = None


class Extraction(BaseModel):
    id: str = Field(default_factory=lambda: "extr_" + nanoid.generate(), description="Unique identifier of the analysis")
    messages: list[ChatCompletionUiformMessage] = Field(default_factory=list)
    messages_gcs: str = Field(..., description="GCS path to the messages")
    file_gcs: str = Field(..., description="GCS path to the file")
    file_id: str = Field(..., description="ID of the file")

    status: Literal["success", "failed"] = Field(..., description="Whether the analysis was successful")
    completion: UiParsedChatCompletion | ChatCompletion = Field(..., description="Response generated by the analysis")
    json_schema: Any = Field(..., description="Response format (JSON Schema or pydantic_v2.BaseModel)")
    model: str = Field(..., description="Model used for the analysis")
    temperature: float = Field(default=0.0, description="Temperature used for the analysis")
    source: ExtractionSource = Field(..., description="Source of the extraction")
    image_resolution_dpi: int = Field(default=96, description="Resolution of the image sent to the LLM")
    browser_canvas: Literal['A3', 'A4', 'A5'] = Field(default='A4', description="Sets the size of the browser canvas for rendering documents in browser-based processing. Choose a size that matches the document type.")
    modality: Modality = Field(default="native", description="Modality of the extraction")
    reasoning_effort: Optional[ChatCompletionReasoningEffort] = Field(default=None, description="The effort level for the model to reason about the input data.")
    timings: list[ExtractionTimingStep] = Field(default_factory=list, description="Timings of the extraction")

    # Infered from the schema
    schema_id: str = Field(..., description="Version of the schema used for the analysis")
    schema_data_id: str = Field(..., description="Version of the schema data used for the analysis")
    created_at: datetime.datetime = Field(default_factory=lambda: datetime.datetime.now(datetime.timezone.utc), description="Timestamp of the creation of the extraction object")
    request_at: datetime.datetime | None = Field(default=None, description="Timestamp of the extraction request if provided.")
    organization_id: str = Field(..., description="Organization ID of the user or application")
    validation_state: Optional[ValidationsState] = Field(default=None, description="Validation state of the extraction")
    billed: bool = Field(default=False, description="Whether the extraction has been billed or not")

    @computed_field
    @property
    def api_cost(self) -> Optional[Amount]:
        if self.completion and self.completion.usage:
            try:
                cost = compute_cost_from_model(self.completion.model, self.completion.usage)
                return cost
            except Exception as e:
                print(f"Error computing cost: {e}")
                return None
        return None
    
    @computed_field  # type: ignore
    @property
    def cost_breakdown(self) -> Optional[CostBreakdown]:
        if self.completion and self.completion.usage:
            try:
                cost = compute_cost_from_model_with_breakdown(self.completion.model, self.completion.usage)
                return cost
            except Exception as e:
                print(f"Error computing cost: {e}")
                return None
        return None



import base64
import json
from io import IOBase
from pathlib import Path
from typing import Any, AsyncGenerator, Generator, Literal, Optional

from anthropic.types.message_param import MessageParam
from openai.types.chat import ChatCompletionMessageParam
from openai.types.chat.chat_completion_reasoning_effort import ChatCompletionReasoningEffort
from openai.types.chat.parsed_chat_completion import ParsedChatCompletionMessage
from openai.types.responses.response import Response
from openai.types.responses.response_input_param import ResponseInputItemParam
from pydantic import HttpUrl

from ..._resource import AsyncAPIResource, SyncAPIResource
from ..._utils.ai_models import assert_valid_model_extraction
from ..._utils.json_schema import filter_auxiliary_fields_json, load_json_schema, unflatten_dict
from ..._utils.mime import MIMEData, prepare_mime_document
from ..._utils.stream_context_managers import as_async_context_manager, as_context_manager
from ...types.chat import ChatCompletionUiformMessage
from ...types.documents.extractions import DocumentExtractRequest, LogExtractionRequest, UiParsedChatCompletion, UiParsedChatCompletionChunk, UiParsedChoice
from ...types.modalities import Modality
from ...types.schemas.object import Schema
from ...types.standards import PreparedRequest


def maybe_parse_to_pydantic(schema: Schema, response: UiParsedChatCompletion, allow_partial: bool = False) -> UiParsedChatCompletion:
    if response.choices[0].message.content:
        try:
            if allow_partial:
                response.choices[0].message.parsed = schema._partial_pydantic_model.model_validate(filter_auxiliary_fields_json(response.choices[0].message.content))
            else:
                response.choices[0].message.parsed = schema.pydantic_model.model_validate(filter_auxiliary_fields_json(response.choices[0].message.content))
        except Exception as e:
            pass
    return response


class BaseExtractionsMixin:
    def prepare_extraction(
        self,
        json_schema: dict[str, Any] | Path | str,
        document: Path | str | IOBase | HttpUrl | None,
        image_resolution_dpi: int | None,
        browser_canvas: Literal['A3', 'A4', 'A5'] | None,
        model: str,
        temperature: float,
        modality: Modality,
        reasoning_effort: ChatCompletionReasoningEffort,
        stream: bool,
        n_consensus: int = 1,
        store: bool = False,
        idempotency_key: str | None = None,
    ) -> PreparedRequest:
        assert_valid_model_extraction(model)

        json_schema = load_json_schema(json_schema)

        data = {
            "json_schema": json_schema,
            "document": prepare_mime_document(document).model_dump() if document is not None else None,
            "model": model,
            "temperature": temperature,
            "stream": stream,
            "modality": modality,
            "store": store,
            "reasoning_effort": reasoning_effort,
            "n_consensus": n_consensus,
        }
        if image_resolution_dpi:
            data["image_resolution_dpi"] = image_resolution_dpi
        if browser_canvas:
            data["browser_canvas"] = browser_canvas

        # Validate DocumentAPIRequest data (raises exception if invalid)
        document_extract_request = DocumentExtractRequest.model_validate(data)

        return PreparedRequest(method="POST", url="/v1/documents/extractions", data=document_extract_request.model_dump(), idempotency_key=idempotency_key)

    def prepare_log_extraction(
        self,
        document: Path | str | IOBase | HttpUrl | None,
        json_schema: dict[str, Any],
        model: str,
        temperature: float,
        completion: Any | None = None,
        # The messages can be provided in different formats, we will convert them to the UiForm-compatible format
        messages: list[ChatCompletionUiformMessage] | None = None,
        openai_messages: list[ChatCompletionMessageParam] | None = None,
        anthropic_messages: list[MessageParam] | None = None,
        anthropic_system_prompt: str | None = None,
        # New fields for the Responses API
        openai_responses_input: list[ResponseInputItemParam] | None = None,
        openai_responses_output: Response | None = None,
    ) -> PreparedRequest:
        if document is None:
            mime_document = MIMEData(
                filename="dummy.txt",
                # url is a base64 encoded string with the mime type and the content. For the dummy one we will send a .txt file with the text "No document provided"
                url="data:text/plain;base64," + base64.b64encode(b"No document provided").decode("utf-8"),
            )

        return PreparedRequest(
            method="POST",
            url="/v1/documents/log_extraction",
            data=LogExtractionRequest(
                document=prepare_mime_document(document) if document else mime_document,
                messages=messages,
                openai_messages=openai_messages,
                anthropic_messages=anthropic_messages,
                anthropic_system_prompt=anthropic_system_prompt,
                completion=completion,
                openai_responses_input=openai_responses_input,
                openai_responses_output=openai_responses_output,
                json_schema=json_schema,
                model=model,
                temperature=temperature,
            ).model_dump(mode="json", by_alias=True),  # by_alias is necessary to enable serialization/deserialization ('schema' was being converted to 'schema_')
            raise_for_status=True,
        )


class Extractions(SyncAPIResource, BaseExtractionsMixin):
    """Extraction API wrapper"""

    def parse(
        self,
        json_schema: dict[str, Any] | Path | str,
        model: str,
        document: Path | str | IOBase | HttpUrl | None,
        image_resolution_dpi: int | None = None,
        browser_canvas: Literal['A3', 'A4', 'A5'] | None = None,
        temperature: float = 0,
        modality: Modality = "native",
        reasoning_effort: ChatCompletionReasoningEffort = "medium",
        n_consensus: int = 1,
        idempotency_key: str | None = None,
        store: bool = False,
    ) -> UiParsedChatCompletion:
        """
        Process a document using the UiForm API.

        Args:
            json_schema: JSON schema defining the expected data structure
            document: Single document (as MIMEData) to process
            model: The AI model to use for processing
            temperature: Model temperature setting (0-1)
            modality: Modality of the document (e.g., native)
            reasoning_effort: The effort level for the model to reason about the input data.
            n_consensus: Number of consensus extractions to perform (default: 1 which computes a single extraction and the likelihoods comes from the model logprobs)
            idempotency_key: Idempotency key for request
            store: Whether to store the document in the UiForm database
        Returns:
            DocumentAPIResponse
        Raises:
            HTTPException if the request fails
        """

        assert document is not None, "Either document or messages must be provided"

        # Validate DocumentAPIRequest data (raises exception if invalid)
        request = self.prepare_extraction(
            json_schema, document, image_resolution_dpi, browser_canvas, model, temperature, modality, reasoning_effort, False, n_consensus=n_consensus, store=store, idempotency_key=idempotency_key
        )
        response = self._client._prepared_request(request)

        schema = Schema(json_schema=load_json_schema(json_schema))
        return maybe_parse_to_pydantic(schema, UiParsedChatCompletion.model_validate(response))

    @as_context_manager
    def stream(
        self,
        json_schema: dict[str, Any] | Path | str,
        model: str,
        document: Path | str | IOBase | HttpUrl | None,
        image_resolution_dpi: int | None = None,
        browser_canvas: Literal['A3', 'A4', 'A5'] | None = None,
        temperature: float = 0,
        modality: Modality = "native",
        reasoning_effort: ChatCompletionReasoningEffort = "medium",
        n_consensus: int = 1,
        idempotency_key: str | None = None,
        store: bool = False,
    ) -> Generator[UiParsedChatCompletion, None, None]:
        """
        Process a document using the UiForm API with streaming enabled.

        Args:
            json_schema: JSON schema defining the expected data structure
            document: Single document (as MIMEData) to process
            image_resolution_dpi: Optional image resolution DPI.
            browser_canvas: Optional browser canvas size.
            model: The AI model to use for processing
            temperature: Model temperature setting (0-1)
            modality: Modality of the document (e.g., native)
            reasoning_effort: The effort level for the model to reason about the input data.
            n_consensus: Number of consensus extractions to perform (default: 1 which computes a single extraction and the likelihoods comes from the model logprobs)
            idempotency_key: Idempotency key for request
            store: Whether to store the document in the UiForm database

        Returns:
            Generator[DocumentExtractResponse]: Stream of parsed responses
        Raises:
            HTTPException if the request fails
        Usage:
        ```python
        with uiform.documents.extractions.stream(json_schema, document, model, temperature, reasoning_effort, modality) as stream:
            for response in stream:
                print(response)
        ```
        """
        request = self.prepare_extraction(
            json_schema, document, image_resolution_dpi, browser_canvas, model, temperature, modality, reasoning_effort, True, n_consensus=n_consensus, store=store, idempotency_key=idempotency_key
        )
        schema = Schema(json_schema=load_json_schema(json_schema))

        # Request the stream and return a context manager
        ui_parsed_chat_completion_cum_chunk: UiParsedChatCompletionChunk | None = None
        # Initialize the UiParsedChatCompletion object
        ui_parsed_completion: UiParsedChatCompletion = UiParsedChatCompletion(
            id="",
            created=0,
            model="",
            object="chat.completion",
            likelihoods={},
            choices=[
                UiParsedChoice(
                    index=0,
                    message=ParsedChatCompletionMessage(content="", role="assistant"),
                    finish_reason=None,
                    logprobs=None,
                )
            ],
        )
        for chunk_json in self._client._prepared_request_stream(request):
            if not chunk_json:
                continue
            ui_parsed_chat_completion_cum_chunk = UiParsedChatCompletionChunk.model_validate(chunk_json).chunk_accumulator(ui_parsed_chat_completion_cum_chunk)
            # Basic stuff
            ui_parsed_completion.id = ui_parsed_chat_completion_cum_chunk.id
            ui_parsed_completion.created = ui_parsed_chat_completion_cum_chunk.created
            ui_parsed_completion.model = ui_parsed_chat_completion_cum_chunk.model
            # Update the ui_parsed_completion object
            parsed = unflatten_dict(ui_parsed_chat_completion_cum_chunk.choices[0].delta.flat_parsed)
            likelihoods = unflatten_dict(ui_parsed_chat_completion_cum_chunk.choices[0].delta.flat_likelihoods)
            ui_parsed_completion.choices[0].message.content = json.dumps(parsed)
            ui_parsed_completion.choices[0].message.parsed = parsed
            ui_parsed_completion.likelihoods = likelihoods

            yield maybe_parse_to_pydantic(schema, ui_parsed_completion, allow_partial=True)

        # change the finish_reason to stop
        ui_parsed_completion.choices[0].finish_reason = "stop"
        yield maybe_parse_to_pydantic(schema, ui_parsed_completion)

    def log(
        self,
        document: Path | str | IOBase | HttpUrl | None,
        json_schema: dict[str, Any],
        model: str,
        temperature: float,
        completion: Any | None = None,
        # The messages can be provided in different formats, we will convert them to the UiForm-compatible format
        messages: list[ChatCompletionUiformMessage] | None = None,
        openai_messages: list[ChatCompletionMessageParam] | None = None,
        anthropic_messages: list[MessageParam] | None = None,
        anthropic_system_prompt: str | None = None,
        # New fields for the Responses API
        openai_responses_input: list[ResponseInputItemParam] | None = None,
        openai_responses_output: Response | None = None,
    ) -> None:
        request = self.prepare_log_extraction(
            document,
            json_schema,
            model,
            temperature,
            completion=completion,
            messages=messages,
            openai_messages=openai_messages,
            anthropic_messages=anthropic_messages,
            anthropic_system_prompt=anthropic_system_prompt,
            openai_responses_input=openai_responses_input,
            openai_responses_output=openai_responses_output,
        )
        return self._client._prepared_request(request)


class AsyncExtractions(AsyncAPIResource, BaseExtractionsMixin):
    """Extraction API wrapper for asynchronous usage."""

    async def parse(
        self,
        json_schema: dict[str, Any] | Path | str,
        model: str,
        document: Path | str | IOBase | HttpUrl | None,
        image_resolution_dpi: int | None = None,
        browser_canvas: Literal['A3', 'A4', 'A5'] | None = None,
        temperature: float = 0,
        modality: Modality = "native",
        reasoning_effort: ChatCompletionReasoningEffort = "medium",
        n_consensus: int = 1,
        idempotency_key: str | None = None,
        store: bool = False,
    ) -> UiParsedChatCompletion:
        """
        Extract structured data from a document asynchronously.

        Args:
            json_schema: JSON schema defining the expected data structure.
            document: Path, string, or file-like object representing the document.
            image_resolution_dpi: Optional image resolution DPI.
            browser_canvas: Optional browser canvas size.
            model: The AI model to use.
            temperature: Model temperature setting (0-1).
            modality: Modality of the document (e.g., native).
            reasoning_effort: The effort level for the model to reason about the input data.
            n_consensus: Number of consensus extractions to perform (default: 1 which computes a single extraction and the likelihoods comes from the model logprobs)
            idempotency_key: Idempotency key for request
            store: Whether to store the document in the UiForm database
        Returns:
            DocumentExtractResponse: Parsed response from the API.
        """
        request = self.prepare_extraction(
            json_schema, document, image_resolution_dpi, browser_canvas, model, temperature, modality, reasoning_effort, False, n_consensus=n_consensus, store=store, idempotency_key=idempotency_key
        )
        response = await self._client._prepared_request(request)
        schema = Schema(json_schema=load_json_schema(json_schema))
        return maybe_parse_to_pydantic(schema, UiParsedChatCompletion.model_validate(response))

    @as_async_context_manager
    async def stream(
        self,
        json_schema: dict[str, Any] | Path | str,
        model: str,
        document: Path | str | IOBase | HttpUrl | None,
        image_resolution_dpi: int | None = None,
        browser_canvas: Literal['A3', 'A4', 'A5'] | None = None,
        temperature: float = 0,
        modality: Modality = "native",
        reasoning_effort: ChatCompletionReasoningEffort = "medium",
        n_consensus: int = 1,
        idempotency_key: str | None = None,
        store: bool = False,
    ) -> AsyncGenerator[UiParsedChatCompletion, None]:
        """
        Extract structured data from a document asynchronously with streaming.

        Args:
            json_schema: JSON schema defining the expected data structure.
            document: Path, string, or file-like object representing the document.
            model: The AI model to use.
            temperature: Model temperature setting (0-1).
            modality: Modality of the document (e.g., native).
            reasoning_effort: The effort level for the model to reason about the input data.
            n_consensus: Number of consensus extractions to perform (default: 1 which computes a single extraction and the likelihoods comes from the model logprobs)
            idempotency_key: Idempotency key for request
            store: Whether to store the document in the UiForm database
        Returns:
            AsyncGenerator[DocumentExtractResponse, None]: Stream of parsed responses.

        Usage:
        ```python
        async with uiform.documents.extractions.stream(json_schema, document, model, temperature, reasoning_effort, modality) as stream:
            async for response in stream:
                print(response)
        ```
        """
        request = self.prepare_extraction(
            json_schema, document, image_resolution_dpi, browser_canvas, model, temperature, modality, reasoning_effort, True, n_consensus=n_consensus, store=store, idempotency_key=idempotency_key
        )
        schema = Schema(json_schema=load_json_schema(json_schema))
        ui_parsed_chat_completion_cum_chunk: UiParsedChatCompletionChunk | None = None
        # Initialize the UiParsedChatCompletion object
        ui_parsed_completion: UiParsedChatCompletion = UiParsedChatCompletion(
            id="",
            created=0,
            model="",
            object="chat.completion",
            likelihoods={},
            choices=[
                UiParsedChoice(
                    index=0,
                    message=ParsedChatCompletionMessage(content="", role="assistant"),
                    finish_reason=None,
                    logprobs=None,
                )
            ],
        )

        async for chunk_json in self._client._prepared_request_stream(request):
            if not chunk_json:
                continue
            ui_parsed_chat_completion_cum_chunk = UiParsedChatCompletionChunk.model_validate(chunk_json).chunk_accumulator(ui_parsed_chat_completion_cum_chunk)
            # Basic stuff
            ui_parsed_completion.id = ui_parsed_chat_completion_cum_chunk.id
            ui_parsed_completion.created = ui_parsed_chat_completion_cum_chunk.created
            ui_parsed_completion.model = ui_parsed_chat_completion_cum_chunk.model

            # Update the ui_parsed_completion object
            parsed = unflatten_dict(ui_parsed_chat_completion_cum_chunk.choices[0].delta.flat_parsed)
            likelihoods = unflatten_dict(ui_parsed_chat_completion_cum_chunk.choices[0].delta.flat_likelihoods)
            ui_parsed_completion.choices[0].message.content = json.dumps(parsed)
            ui_parsed_completion.choices[0].message.parsed = parsed
            ui_parsed_completion.likelihoods = likelihoods

            yield maybe_parse_to_pydantic(schema, ui_parsed_completion, allow_partial=True)

        # change the finish_reason to stop
        ui_parsed_completion.choices[0].finish_reason = "stop"
        yield maybe_parse_to_pydantic(schema, ui_parsed_completion)

    async def log(
        self,
        document: Path | str | IOBase | HttpUrl | None,
        json_schema: dict[str, Any],
        model: str,
        temperature: float,
        completion: Any | None = None,
        # The messages can be provided in different formats, we will convert them to the UiForm-compatible format
        messages: list[ChatCompletionUiformMessage] | None = None,
        openai_messages: list[ChatCompletionMessageParam] | None = None,
        anthropic_messages: list[MessageParam] | None = None,
        anthropic_system_prompt: str | None = None,
        # New fields for the Responses API
        openai_responses_input: list[ResponseInputItemParam] | None = None,
        openai_responses_output: Response | None = None,
    ) -> None:
        request = self.prepare_log_extraction(
            document,
            json_schema,
            model,
            temperature,
            completion=completion,
            messages=messages,
            openai_messages=openai_messages,
            anthropic_messages=anthropic_messages,
            anthropic_system_prompt=anthropic_system_prompt,
            openai_responses_input=openai_responses_input,
            openai_responses_output=openai_responses_output,
        )
        return await self._client._prepared_request(request)



import datetime
import json
from io import IOBase
from pathlib import Path
from typing import Any, Dict, Literal, Optional

import httpx
from pydantic import HttpUrl
from PIL.Image import Image

from ...._resource import AsyncAPIResource, SyncAPIResource
from ...._utils.mime import prepare_mime_document
from ....types.mime import BaseMIMEData, MIMEData
from ....types.logs import AutomationLog
from ....types.standards import PreparedRequest


class TestsMixin:
    def prepare_upload(self, automation_id: str, document: Path | str | IOBase | HttpUrl | Image | MIMEData) -> PreparedRequest:
        mime_document = prepare_mime_document(document)
        return PreparedRequest(method="POST", url=f"/v1/processors/automations/tests/upload/{automation_id}", data={"document": mime_document.model_dump(mode='json')})

    def prepare_webhook(self, automation_id: str) -> PreparedRequest:
        return PreparedRequest(method="POST", url=f"/v1/processors/automations/tests/webhook/{automation_id}", data=None)

    def print_upload_verbose(self, log: AutomationLog) -> None:
        if log.external_request_log:
            print(f"\nTEST FILE UPLOAD RESULTS:")
            print(f"\n#########################")
            print(f"Status Code: {log.external_request_log.status_code}")
            print(f"Duration: {log.external_request_log.duration_ms:.2f}ms")

            if log.external_request_log.error:
                print(f"\nERROR: {log.external_request_log.error}")

            if log.external_request_log.response_body:
                print("\n--------------")
                print("RESPONSE BODY:")
                print("--------------")
                print(json.dumps(log.external_request_log.response_body, indent=2))
                
            if log.external_request_log.response_headers:
                print("\n--------------")
                print("RESPONSE HEADERS:")
                print("--------------")
                print(json.dumps(log.external_request_log.response_headers, indent=2))

    def print_webhook_verbose(self, log: AutomationLog) -> None:
        if log.external_request_log:
            print(f"\nTEST WEBHOOK RESULTS:")
            print(f"\n#########################")
            print(f"Status Code: {log.external_request_log.status_code}")
            print(f"Duration: {log.external_request_log.duration_ms:.2f}ms")

            if log.external_request_log.error:
                print(f"\nERROR: {log.external_request_log.error}")

            if log.external_request_log.response_body:
                print("\n--------------")
                print("RESPONSE BODY:")
                print("--------------")
                print(json.dumps(log.external_request_log.response_body, indent=2))
                
            if log.external_request_log.response_headers:
                print("\n--------------")
                print("RESPONSE HEADERS:")
                print("--------------")
                print(json.dumps(log.external_request_log.response_headers, indent=2))


class Tests(SyncAPIResource, TestsMixin):
    """Test API wrapper for testing automation configurations"""

    def upload(self, automation_id: str, document: Path | str | IOBase | HttpUrl | Image | MIMEData, verbose: bool = True) -> AutomationLog:
        """Test endpoint that simulates the complete extraction process with the provided document.

        Args:
            automation_id: ID of the automation to test
            document: Document to process
            verbose: Whether to print verbose output

        Returns:
            AutomationLog: The automation log with extraction results
        """
        request = self.prepare_upload(automation_id, document)
        response = self._client._prepared_request(request)
        
        log = AutomationLog.model_validate(response)
        
        if verbose:
            self.print_upload_verbose(log)
            
        return log

    def webhook(self, automation_id: str, verbose: bool = True) -> AutomationLog:
        """Test endpoint that simulates the complete webhook process with sample data.

        Args:    
            automation_id: ID of the automation to test
            verbose: Whether to print verbose output

        Returns:
            AutomationLog: The automation log with webhook results
        """
        request = self.prepare_webhook(automation_id)
        response = self._client._prepared_request(request)
        
        log = AutomationLog.model_validate(response)
        
        if verbose:
            self.print_webhook_verbose(log)
            
        return log


class AsyncTests(AsyncAPIResource, TestsMixin):
    """Async Test API wrapper for testing deployment configurations"""

    async def upload(self, automation_id: str, document: Path | str | IOBase | HttpUrl | Image | MIMEData, verbose: bool = True) -> AutomationLog:
        """Test endpoint that simulates the complete extraction process with the provided document.

        Args:
            automation_id: ID of the automation to test
            document: Document to process
            verbose: Whether to print verbose output

        Returns:
            AutomationLog: The automation log with extraction results
        """
        request = self.prepare_upload(automation_id, document)
        response = await self._client._prepared_request(request)
        
        log = AutomationLog.model_validate(response)
        
        if verbose:
            self.print_upload_verbose(log)
            
        return log

    async def webhook(self, automation_id: str, verbose: bool = True) -> AutomationLog:
        """Test endpoint that simulates the complete webhook process with sample data.

        Args:
            automation_id: ID of the automation to test
            verbose: Whether to print verbose output

        Returns:
            AutomationLog: The automation log with webhook results
        """
        request = self.prepare_webhook(automation_id)
        response = await self._client._prepared_request(request)
        
        log = AutomationLog.model_validate(response)
        
        if verbose:
            self.print_webhook_verbose(log)
            
        return log



import asyncio
import hashlib
import json
import re
import shutil
import tempfile
import time
from concurrent.futures import ThreadPoolExecutor
from io import IOBase
from pathlib import Path
from typing import IO, Any, Literal, Optional

from anthropic import Anthropic
from openai import OpenAI
from openai.types.chat.chat_completion_message_param import ChatCompletionMessageParam
from pydantic import BaseModel
from tqdm import tqdm

from .._resource import AsyncAPIResource, SyncAPIResource
from .._utils.ai_models import assert_valid_model_extraction, find_provider_from_model
from .._utils.chat import convert_to_anthropic_format, convert_to_openai_format, separate_messages
from .._utils.display import Metrics, display_metrics, process_dataset_and_compute_metrics
from .._utils.json_schema import load_json_schema
from ..types.chat import ChatCompletionUiformMessage
from ..types.modalities import Modality
from ..types.schemas.object import Schema


class FinetuningJSON(BaseModel):
    messages: list[ChatCompletionUiformMessage]


FinetuningJSONL = list[FinetuningJSON]
from typing import TypedDict


class BatchJSONLResponseFormat(TypedDict):
    type: str
    json_schema: dict[str, Any]


class BatchJSONLBody(TypedDict):
    model: str
    messages: list[ChatCompletionMessageParam]
    temperature: float
    response_format: BatchJSONLResponseFormat


class BatchJSONL(TypedDict):
    custom_id: str
    method: str
    url: str
    body: BatchJSONLBody


class BatchJSONLResponseUsageTokenDetails(BaseModel):
    cached_tokens: int
    audio_tokens: int


class BatchJSONLResponseUsageCompletionDetails(BaseModel):
    reasoning_tokens: int
    audio_tokens: int
    accepted_prediction_tokens: int
    rejected_prediction_tokens: int


class BatchJSONLResponseUsage(BaseModel):
    prompt_tokens: int
    completion_tokens: int
    total_tokens: int
    prompt_tokens_details: BatchJSONLResponseUsageTokenDetails
    completion_tokens_details: BatchJSONLResponseUsageCompletionDetails


class BatchJSONLResponseChoice(BaseModel):
    index: int
    message: ChatCompletionMessageParam
    logprobs: None | Any
    finish_reason: str


class BatchJSONLResponseBody(BaseModel):
    id: str
    object: str
    created: int
    model: str
    choices: list[BatchJSONLResponseChoice]
    usage: BatchJSONLResponseUsage
    service_tier: str
    system_fingerprint: str


class BatchJSONLResponseInner(BaseModel):
    status_code: int
    request_id: str
    body: BatchJSONLResponseBody


class BatchJSONLResponse(BaseModel):
    id: str
    custom_id: str
    response: BatchJSONLResponseInner
    error: None | str


class BaseDatasetsMixin:
    def _dump_training_set(self, training_set: list[dict[str, Any]], dataset_path: Path | str) -> None:
        with open(dataset_path, 'w', encoding='utf-8') as file:
            for entry in training_set:
                file.write(json.dumps(entry) + '\n')


class Datasets(SyncAPIResource, BaseDatasetsMixin):
    """Datasets API wrapper"""

    # TODO : Maybe at some point we could add some visualization methods... but the multimodality makes it hard... # client.datasets.plot.tsne()... # client.datasets.plot.umap()...
    def pprint(self, dataset_path: Path | str, input_token_price: Optional[float] = None, output_token_price: Optional[float] = None) -> Metrics:
        """Print a summary of the contents and statistics of a JSONL file.

        This method analyzes the JSONL file and displays various metrics and statistics
        about the dataset contents.

        Inspired from : https://gist.github.com/nmwsharp/54d04af87872a4988809f128e1a1d233

        Args:
            dataset_path: Path to the JSONL file to analyze
            completion_token_price: Price per completion token
            input_token_price: Price per input token
        """

        computed_metrics = process_dataset_and_compute_metrics(dataset_path)
        display_metrics(computed_metrics, input_token_price=input_token_price, output_token_price=output_token_price)
        return computed_metrics

    def save(
        self,
        json_schema: dict[str, Any] | Path | str,
        document_annotation_pairs_paths: list[dict[str, Path | str]],
        dataset_path: Path | str,
        image_resolution_dpi: int | None = None,
        browser_canvas: Literal['A3', 'A4', 'A5'] | None = None,
        modality: Modality = "native",
    ) -> None:
        """Save document-annotation pairs to a JSONL training set.

        Args:
            json_schema: The JSON schema for validation, can be a dict, Path, or string
            document_annotation_pairs_paths: {document_fpath: Path | str, annotation_fpath: Path | str} List of dictionaries containing document and annotation file paths
            jsonl_path: Output path for the JSONL training file
            modality: The modality to use for document processing ("native" by default)
        """
        json_schema = load_json_schema(json_schema)
        schema_obj = Schema(json_schema=json_schema)

        with open(dataset_path, 'w', encoding='utf-8') as file:
            for pair_paths in tqdm(document_annotation_pairs_paths, desc="Processing pairs", position=0):
                document_message = self._client.documents.create_messages(document=pair_paths['document_fpath'], modality=modality, image_resolution_dpi=image_resolution_dpi, browser_canvas=browser_canvas)

                with open(pair_paths['annotation_fpath'], 'r') as f:
                    annotation = json.loads(f.read())
                assistant_message = {"role": "assistant", "content": json.dumps(annotation, ensure_ascii=False, indent=2)}

                entry = {"messages": schema_obj.messages + document_message.messages + [assistant_message]}
                file.write(json.dumps(entry) + '\n')

    def change_schema(
        self,
        input_dataset_path: Path | str,
        json_schema: dict[str, Any] | Path | str,
        output_dataset_path: None | Path | str = None,
        inplace: bool = False,
    ) -> None:
        """Change the system prompt in a dataset to match a new schema.

        Args:
            input_dataset_path: Path to the input JSONL dataset file
            output_dataset_path: Path to the output JSONL dataset file
            json_schema: The new JSON schema for validation, can be a dict, Path, or string
            inplace: If True, overwrite the input dataset with the changes
        """
        json_schema = load_json_schema(json_schema)
        schema_obj = Schema(json_schema=json_schema)

        # Determine the path to write to
        if inplace:
            assert output_dataset_path is None, "Cannot specify inplace=True and output_dataset_path not None"
            target_path = input_dataset_path
        else:
            assert output_dataset_path, "Cannot save the file if inplace=False and output_dataset_path is None"
            target_path = output_dataset_path

        assert isinstance(target_path, Path) or isinstance(target_path, str)

        # Use a temporary file to write the updated content
        with tempfile.NamedTemporaryFile('w', delete=False, encoding='utf-8') as temp_file:
            with open(input_dataset_path, 'r', encoding='utf-8') as infile:
                for line in infile:
                    entry = json.loads(line)
                    messages = entry.get("messages", [])

                    # Remove existing system prompt if it exists
                    if messages and messages[0].get("role") in ("developer", "system"):
                        messages = messages[1:]

                    # Add the new system prompt from schema_obj.messages
                    updated_messages = schema_obj.messages + messages
                    updated_entry = {"messages": updated_messages}

                    # Write the updated entry to the temporary file
                    temp_file.write(json.dumps(updated_entry) + '\n')

        # Replace the original file with the temporary file
        shutil.move(temp_file.name, target_path)

    def stich_and_save(
        self,
        json_schema: dict[str, Any] | Path | str,
        pairs_paths: list[dict[str, Path | str | list[Path | str] | list[str] | list[Path]]],
        dataset_path: Path | str,
        modality: Modality = "native",
    ) -> None:
        """Stitch multiple documents and their annotations into a JSONL training set.

        This method processes and combines multiple documents into messages, creating document-annotation
        pairs that are saved to a JSONL file. Each document is processed according to the specified
        modality and combined with its corresponding annotation.


        Args:
            json_schema: The JSON schema for validation, can be a dict, Path, or string
            pairs_paths: List of dictionaries containing document and annotation file paths
            jsonl_path: Output path for the JSONL training file
            modality: The modality to use for document processing ("native" by default)
        """

        json_schema = load_json_schema(json_schema)
        schema_obj = Schema(json_schema=json_schema)
        training_set = []

        for pair_paths in tqdm(pairs_paths):
            document_messages: list[ChatCompletionUiformMessage] = []

            if isinstance(pair_paths['document_fpath'], str) or isinstance(pair_paths['document_fpath'], Path):
                document_message = self._client.documents.create_messages(document=pair_paths['document_fpath'], modality=modality)
                document_messages.extend(document_message.messages)

            else:
                assert isinstance(pair_paths['document_fpath'], list)
                for document_fpath in pair_paths['document_fpath']:
                    document_message = self._client.documents.create_messages(document=document_fpath, modality=modality)
                    document_messages.extend(document_message.messages)

            # Use context manager to properly close the file
            assert isinstance(pair_paths['annotation_fpath'], Path) or isinstance(pair_paths['annotation_fpath'], str)
            with open(pair_paths['annotation_fpath'], 'r') as f:
                annotation = json.loads(f.read())
            assistant_message = {"role": "assistant", "content": json.dumps(annotation, ensure_ascii=False, indent=2)}

            # Add the complete message set as an entry
            training_set.append({"messages": schema_obj.messages + document_messages + [assistant_message]})

        self._dump_training_set(training_set, dataset_path)

    #########################################
    ##### ENDPOINTS THAT MAKE LLM CALLS #####
    #########################################

    def _initialize_model_client(self, model: str) -> tuple[OpenAI | Anthropic, str]:
        """Initialize the appropriate client based on the model provider.

        Args:
            model: The model identifier string

        Returns:
            A tuple of (client instance, provider type string)
        """
        provider = find_provider_from_model(model)

        if provider == "OpenAI":
            return OpenAI(api_key=self._client.headers["OpenAI-Api-Key"]), provider
        elif provider == "xAI":
            return OpenAI(api_key=self._client.headers["XAI-Api-Key"], base_url="https://api.x.ai/v1"), provider
        elif provider == "Gemini":
            return OpenAI(
                api_key=self._client.headers["Gemini-Api-Key"],
                base_url="https://generativelanguage.googleapis.com/v1beta/openai/",
            ), provider
        else:
            assert provider == "Anthropic", f"Unsupported model: {model}"
            return Anthropic(api_key=self._client.headers["Anthropic-Api-Key"]), provider

    def _get_model_completion(
        self,
        client: OpenAI | Anthropic,
        provider: str,
        model: str,
        temperature: float,
        messages: list[ChatCompletionUiformMessage],
        schema_obj: Schema,
    ) -> str:
        """Get completion from the appropriate model provider.

        Args:
            client: The initialized client instance
            provider: The provider type string
            model: The model identifier
            temperature: Temperature setting for generation
            messages: The messages to send to the model
            schema_obj: The schema object containing format information

        Returns:
            The completion string in JSON format
        """
        if provider in ["OpenAI", "xAI"]:
            assert isinstance(client, OpenAI)
            completion = client.chat.completions.create(
                model=model,
                temperature=temperature,
                messages=convert_to_openai_format(messages),
                response_format={"type": "json_schema", "json_schema": {"name": schema_obj.id, "schema": schema_obj.inference_json_schema, "strict": True}},
            )
            assert completion.choices[0].message.content is not None
            return completion.choices[0].message.content

        elif provider == "Gemini":
            assert isinstance(client, OpenAI)
            gemini_completion = client.chat.completions.create(
                model=model,
                temperature=temperature,
                messages=convert_to_openai_format(messages),
                response_format={"type": "json_schema", "json_schema": {"name": schema_obj.id, "schema": schema_obj.inference_gemini_json_schema, "strict": True}},
            )
            assert gemini_completion.choices[0].message.content is not None
            return gemini_completion.choices[0].message.content

        else:  # Anthropic
            assert isinstance(client, Anthropic)
            system_messages, other_messages = convert_to_anthropic_format(messages)
            anthropic_completion = client.messages.create(
                model="claude-3-5-sonnet-20241022", max_tokens=8192, temperature=temperature, system=system_messages, messages=other_messages
            )
            from anthropic.types.text_block import TextBlock

            assert isinstance(anthropic_completion.content[0], TextBlock)
            return anthropic_completion.content[0].text

    def annotate(
        self,
        json_schema: dict[str, Any] | Path | str,
        documents: list[Path | str | IOBase],
        dataset_path: Path,
        model: str = "gpt-4o-2024-08-06",
        temperature: float = 0.0,
        batch_size: int = 5,
        max_concurrent: int = 3,
        root_dir: Path = Path("annotations"),
        modality: Modality = "native",
    ) -> None:
        json_schema = load_json_schema(json_schema)
        assert_valid_model_extraction(model)

        client, provider = self._initialize_model_client(model)
        schema_obj = Schema(json_schema=json_schema)

        """
        Generate annotations from document files or in-memory documents
        and create a JSONL training set in one go.

        Args:
            json_schema: The JSON schema for validation
            documents: list of documents, each can be a Path/str or an IOBase object
            dataset_path: Output path for the JSONL training file
            model: The model to use for processing
            temperature: Model temperature (0-1)
            batch_size: Number of examples to process in each batch
            max_concurrent: Maximum number of concurrent API calls
            root_dir: Where to store the per-document JSON annotations
        """

        def process_example(doc: Path | str | IOBase) -> dict[str, Any]:
            """
            Process a single document (either a file path or an in-memory file-like object).
            Returns a dict with pointers to the original doc and the stored annotation JSON.
            """
            if isinstance(doc, (str, Path)):
                doc_path = Path(doc)
                if not doc_path.is_file():
                    raise ValueError(f"Invalid file path: {doc_path}")
                hash_str = hashlib.md5(doc_path.as_posix().encode()).hexdigest()
            elif isinstance(doc, IO):
                file_bytes = doc.read()
                hash_str = hashlib.md5(file_bytes).hexdigest()
                doc.seek(0)
            else:
                raise ValueError(f"Unsupported document type: {type(doc)}")

            doc_msg = self._client.documents.create_messages(
                document=doc,
                modality=modality,
            )

            # Use _get_model_completion instead of duplicating provider-specific logic
            string_json = self._get_model_completion(
                client=client, provider=provider, model=model, temperature=temperature, messages=schema_obj.messages + doc_msg.messages, schema_obj=schema_obj
            )

            annotation_path = Path(root_dir) / f"annotations_{hash_str}.json"
            annotation_path.parent.mkdir(parents=True, exist_ok=True)

            with open(annotation_path, 'w', encoding='utf-8') as f:
                json.dump(string_json, f, ensure_ascii=False, indent=2)

            return {"document_fpath": str(doc_path), "annotation_fpath": str(annotation_path)}

        # Make sure output directory exists
        Path(root_dir).mkdir(parents=True, exist_ok=True)

        pairs_paths: list[dict[str, Path | str]] = []
        with ThreadPoolExecutor(max_workers=max_concurrent) as executor:
            futures = []
            # Split documents into batches
            for batch in tqdm([documents[i : i + batch_size] for i in range(0, len(documents), batch_size)], desc="Processing batches"):
                # Submit batch of tasks
                batch_futures = []
                for doc in batch:
                    try:
                        future = executor.submit(process_example, doc)
                        batch_futures.append(future)
                    except Exception as e:
                        print(f"Error submitting document for processing: {e}")
                futures.extend(batch_futures)

                # Wait for batch to finish (rate limit)
                for future in batch_futures:
                    try:
                        pair = future.result()
                        pairs_paths.append(pair)
                    except Exception as e:
                        print(f"Error processing example: {e}")

        # Generate final training set from all results
        self.save(json_schema=json_schema, document_annotation_pairs_paths=pairs_paths, dataset_path=dataset_path)

    def eval(
        self,
        json_schema: dict[str, Any] | Path | str,
        dataset_path: str | Path,
        model: str = "gpt-4o-2024-08-06",
        temperature: float = 0.0,
        batch_size: int = 5,
        max_concurrent: int = 3,
        display: bool = True,
    ) -> ComparisonMetrics:
        """Evaluate model performance on a test dataset.

        Args:
            json_schema: JSON schema defining the expected data structure
            dataset_path: Path to the JSONL file containing test examples
            model: The model to use for benchmarking
            temperature: Model temperature setting (0-1)
            batch_size: Number of examples to process in each batch
            max_concurrent: Maximum number of concurrent API calls
        """

        json_schema = load_json_schema(json_schema)
        assert_valid_model_extraction(model)
        schema_obj = Schema(json_schema=json_schema)

        # Initialize appropriate client
        client, provider = self._initialize_model_client(model)

        # Read all lines from the JSONL file
        with open(dataset_path, 'r') as f:
            lines = [json.loads(line) for line in f]

        extraction_analyses: list[ExtractionAnalysis] = []
        total_batches = (len(lines) + batch_size - 1) // batch_size

        # Create main progress bar for batches
        batch_pbar = tqdm(total=total_batches, desc="Processing batches", position=0)

        # Track running metrics
        class RunningMetrics(BaseModel):
            model: str
            accuracy: float
            levenshtein: float
            jaccard: float
            false_positive: float
            mismatched: float
            processed: int

        running_metrics: RunningMetrics = RunningMetrics(
            model=model,
            accuracy=0.0,
            levenshtein=0.0,
            jaccard=0.0,
            false_positive=0.0,
            mismatched=0.0,
            processed=0,  # number of processed examples - used in the loop to compute the running averages
        )

        def update_running_metrics(analysis: ExtractionAnalysis) -> None:
            comparison = normalized_comparison_metrics([analysis])
            running_metrics.processed += 1
            n = running_metrics.processed
            # Update running averages
            running_metrics.accuracy = (running_metrics.accuracy * (n - 1) + comparison.accuracy) / n
            running_metrics.levenshtein = (running_metrics.levenshtein * (n - 1) + comparison.levenshtein_similarity) / n
            running_metrics.jaccard = (running_metrics.jaccard * (n - 1) + comparison.jaccard_similarity) / n
            running_metrics.false_positive = (running_metrics.false_positive * (n - 1) + comparison.false_positive_rate) / n
            running_metrics.mismatched = (running_metrics.mismatched * (n - 1) + comparison.mismatched_value_rate) / n
            # Update progress bar description
            batch_pbar.set_description(
                f"Processing batches | Model: {running_metrics.model} | Acc: {running_metrics.accuracy:.2f} | "
                f"Lev: {running_metrics.levenshtein:.2f} | "
                f"IOU: {running_metrics.jaccard:.2f} | "
                f"FP: {running_metrics.false_positive:.2f} | "
                f"Mism: {running_metrics.mismatched:.2f}"
            )

        def process_example(jsonline: dict) -> ExtractionAnalysis | None:
            line_number = jsonline['line_number']
            try:
                messages = jsonline['messages']
                ground_truth = json.loads(messages[-1]['content'])
                inference_messages = messages[:-1]

                # Use _get_model_completion instead of duplicating provider-specific logic
                string_json = self._get_model_completion(client=client, provider=provider, model=model, temperature=temperature, messages=inference_messages, schema_obj=schema_obj)

                prediction = json.loads(string_json)
                analysis = ExtractionAnalysis(
                    ground_truth=ground_truth,
                    prediction=prediction,
                )
                update_running_metrics(analysis)
                return analysis
            except Exception as e:
                print(f"\nWarning: Failed to process line number {line_number}: {str(e)}")
                return None

        with ThreadPoolExecutor(max_workers=max_concurrent) as executor:
            # Split entries into batches
            for batch_idx in range(0, len(lines), batch_size):
                batch = lines[batch_idx : batch_idx + batch_size]

                # Submit and process batch
                futures = [executor.submit(process_example, entry | {"line_number": batch_idx * batch_size + i}) for i, entry in enumerate(batch)]
                for future in futures:
                    result = future.result()
                    if result is not None:
                        extraction_analyses.append(result)

                batch_pbar.update(1)

        batch_pbar.close()

        # Analyze error patterns across all examples
        analysis = normalized_comparison_metrics(extraction_analyses)

        if display:
            plot_comparison_metrics(analysis=analysis, top_n=10)

        return analysis

    def benchmark(
        self,
        json_schema: dict[str, Any] | Path | str,
        dataset_path: str | Path,
        models: list[str],
        temperature: float = 0.0,
        batch_size: int = 5,
        max_concurrent: int = 3,
        print: bool = True,
        verbose: bool = False,
    ) -> list[BenchmarkMetrics]:
        """Benchmark multiple models on a test dataset.

        Args:
            json_schema: JSON schema defining the expected data structure
            dataset_path: Path to the JSONL file containing test examples
            models: List of models to benchmark
            temperature: Model temperature setting (0-1)
            batch_size: Number of examples to process in each batch
            max_concurrent: Maximum number of concurrent API calls
            print: Whether to print the metrics
            verbose: Whether to print all the metrics of all the function calls

        Returns:
            Dictionary mapping model names to their evaluation metrics
        """
        results: list[BenchmarkMetrics] = []

        for model in models:
            metrics: ComparisonMetrics = self.eval(
                json_schema=json_schema, dataset_path=dataset_path, model=model, temperature=temperature, batch_size=batch_size, max_concurrent=max_concurrent, display=verbose
            )
            results.append(
                BenchmarkMetrics(
                    ai_model=model,
                    accuracy=metrics.accuracy,
                    levenshtein_similarity=metrics.levenshtein_similarity,
                    jaccard_similarity=metrics.jaccard_similarity,
                    false_positive_rate=metrics.false_positive_rate,
                    false_negative_rate=metrics.false_negative_rate,
                    mismatched_value_rate=metrics.mismatched_value_rate,
                )
            )

        if print:
            display_benchmark_metrics(results)

        return results

    def update_annotations(
        self,
        json_schema: dict[str, Any] | Path | str,
        old_dataset_path: str | Path,
        new_dataset_path: str | Path,
        model: str = "gpt-4o-2024-08-06",
        temperature: float = 0.0,
        batch_size: int = 5,
        max_concurrent: int = 3,
    ) -> None:
        """Update annotations in a JSONL file using a new model.

        Args:
            json_schema: The JSON schema for validation
            old_dataset_path: Path to the JSONL file to update
            new_dataset_path: Path for saving updated annotations
            model: The model to use for new annotations
            temperature: Model temperature (0-1)
            batch_size: Number of examples to process in each batch
            max_concurrent: Maximum number of concurrent API calls
        """
        json_schema = load_json_schema(json_schema)
        assert_valid_model_extraction(model)
        schema_obj = Schema(json_schema=json_schema)

        # Initialize appropriate client
        client, provider = self._initialize_model_client(model)

        # Read all lines from the JSONL file
        with open(old_dataset_path, 'r') as f:
            lines = [json.loads(line) for line in f]

        updated_entries = []
        total_batches = (len(lines) + batch_size - 1) // batch_size

        batch_pbar = tqdm(total=total_batches, desc="Processing batches", position=0)

        def process_entry(entry: dict) -> dict:
            messages = entry['messages']
            system_message, user_messages, assistant_messages = separate_messages(messages)
            system_and_user_messages = messages[:-1]

            previous_annotation_message: ChatCompletionUiformMessage = {
                "role": "user",
                "content": "Here is an old annotation using a different schema. Use it as a reference to update the annotation: " + messages[-1]['content'],
            }

            string_json = self._get_model_completion(
                client=client,
                provider=provider,
                model=model,
                temperature=temperature,
                messages=schema_obj.messages + user_messages + [previous_annotation_message],
                schema_obj=schema_obj,
            )

            return {"messages": system_and_user_messages + [{"role": "assistant", "content": string_json}]}

        with ThreadPoolExecutor(max_workers=max_concurrent) as executor:
            futures = []
            for batch_idx in range(0, len(lines), batch_size):
                batch = lines[batch_idx : batch_idx + batch_size]

                batch_futures = [executor.submit(process_entry, entry | {"line_number": batch_idx * batch_size + i}) for i, entry in enumerate(batch)]
                futures.extend(batch_futures)

                for future in batch_futures:
                    try:
                        result = future.result()
                        updated_entries.append(result)
                    except Exception as e:
                        print(f"Error processing example: {e}")

                batch_pbar.update(1)
                time.sleep(1)

        batch_pbar.close()

        with open(new_dataset_path, 'w') as f:
            for entry in updated_entries:
                f.write(json.dumps(entry) + '\n')

    #########################
    ##### BATCH METHODS #####
    #########################

    def save_batch_annotate_requests(
        self,
        json_schema: dict[str, Any] | Path | str,
        documents: list[Path | str | IOBase],
        batch_requests_path: Path,
        model: str = "gpt-4o-mini",
        temperature: float = 0.0,
        modality: Modality = "native",
    ) -> None:
        """Create a JSONL file containing requests for OpenAI batch processing API.

        Args:
            json_schema: The JSON schema for validation
            documents: List of documents to process
            batch_requests_path: Output path for the JSONL requests file
            model: The model to use for processing
            temperature: Model temperature (0-1)
            modality: The modality to use for document processing
        """
        loaded_json_schema = load_json_schema(json_schema)
        schema_obj = Schema(json_schema=loaded_json_schema)
        assert_valid_model_extraction(model)

        with open(batch_requests_path, 'w', encoding='utf-8') as f:
            for i, doc in tqdm(enumerate(documents)):
                # Create document messages
                doc_msg = self._client.documents.create_messages(
                    document=doc,
                    modality=modality,
                )

                # Construct the request object
                request: BatchJSONL = {
                    "custom_id": f"request-{i}",
                    "method": "POST",
                    "url": "/v1/chat/completions",
                    "body": {
                        "model": model,
                        "messages": schema_obj.openai_messages + doc_msg.openai_messages,
                        "temperature": temperature,
                        "response_format": {"type": "json_schema", "json_schema": {"name": schema_obj.id, "schema": schema_obj.inference_json_schema, "strict": True}},
                    },
                }

                # Write the request as a JSON line
                f.write(json.dumps(request) + '\n')

    def save_batch_update_annotation_requests(
        self,
        json_schema: dict[str, Any] | Path | str,
        old_dataset_path: str | Path,
        batch_requests_path: str | Path,
        model: str = "gpt-4o-mini",
        temperature: float = 0.0,
    ) -> None:
        """Create a JSONL file containing requests for OpenAI batch processing API to update annotations.

        Args:
            json_schema: The JSON schema for validation
            old_dataset_path: Path to the JSONL file to update
            batch_requests_path: Output path for the updated JSONL file
            model: The model to use for processing
            temperature: Model temperature (0-1)
            modality: The modality to use for document processing
        """
        loaded_json_schema = load_json_schema(json_schema)
        schema_obj = Schema(json_schema=loaded_json_schema)

        # Read existing annotations
        with open(old_dataset_path, 'r') as f:
            entries = [json.loads(line) for line in f]

        # Create new JSONL with update requests
        with open(batch_requests_path, 'w', encoding='utf-8') as f:
            for i, entry in enumerate(entries):
                existing_messages = entry['messages']
                system_and_user_messages = existing_messages[:-1]

                previous_annotation_message: ChatCompletionMessageParam = {
                    "role": "user",
                    "content": "Here is an old annotation using a different schema. Use it as a reference to update the annotation: " + existing_messages[-1]['content'],
                }

                # Construct the request object
                response_format: BatchJSONLResponseFormat = {
                    "type": "json_schema",
                    "json_schema": {"name": schema_obj.id, "schema": schema_obj.inference_json_schema, "strict": True},
                }

                body: BatchJSONLBody = {
                    "model": model,
                    "messages": schema_obj.openai_messages + system_and_user_messages + [previous_annotation_message],
                    "temperature": temperature,
                    "response_format": response_format,
                }

                request: BatchJSONL = {"custom_id": f"request-{i}", "method": "POST", "url": "/v1/chat/completions", "body": body}

                # Write the request as a JSON line
                f.write(json.dumps(request) + '\n')

    def build_dataset_from_batch_results(
        self,
        batch_requests_path: str | Path,
        batch_results_path: str | Path,
        dataset_results_path: str | Path,
    ) -> None:
        with open(batch_requests_path, 'r') as f:
            input_lines: list[BatchJSONL] = [json.loads(line) for line in f]
        with open(batch_results_path, 'r') as f:
            batch_results_lines: list[BatchJSONLResponse] = [json.loads(line) for line in f]

        assert len(input_lines) == len(batch_results_lines), "Input and batch results must have the same number of lines"

        for input_line, batch_result in zip(input_lines, batch_results_lines):
            messages = input_line['body']['messages']

            # Filter out messages containing the old annotation reference to remove messages that come from "update annotation"
            if isinstance(messages[-1].get('content'), str):
                if re.search(r'Here is an old annotation using a different schema\. Use it as a reference to update the annotation:', str(messages[-1].get('content', ''))):
                    print("found keyword")
                    input_line['body']['messages'] = messages[:-1]

            input_line['body']['messages'].append(batch_result['response']['body']['choices'][0]['message'])

        with open(dataset_results_path, 'w') as f:
            for input_line in input_lines:
                f.write(json.dumps({'messages': input_line['body']['messages']}) + '\n')

        print(f"Dataset saved to {dataset_results_path}")

    #############################
    ##### END BATCH METHODS #####
    #############################


class AsyncDatasets(AsyncAPIResource, BaseDatasetsMixin):
    """Asynchronous wrapper for Datasets using thread execution."""

    async def save(
        self,
        json_schema: dict[str, Any] | Path | str,
        pairs_paths: list[dict[str, Path | str]],
        jsonl_path: Path | str,
        modality: Modality = "native",
    ) -> None:
        json_schema = load_json_schema(json_schema)
        training_set = []

        for pair_paths in tqdm(pairs_paths):
            document_message = await self._client.documents.create_messages(document=pair_paths['document_fpath'], modality=modality)

            with open(pair_paths['annotation_fpath'], 'r') as f:
                annotation = json.loads(f.read())
            assistant_message = {"role": "assistant", "content": json.dumps(annotation, ensure_ascii=False, indent=2)}

            training_set.append({"messages": document_message.messages + [assistant_message]})

        self._dump_training_set(training_set, jsonl_path)

    async def annotate(
        self,
        json_schema: dict[str, Any] | Path | str,
        documents: list[Path | str | IOBase],
        jsonl_path: str | Path,
        model: str = "gpt-4o-2024-08-06",
        temperature: float = 0.0,
        batch_size: int = 5,
        max_concurrent: int = 3,
        root_dir: Path = Path("annotations"),
        modality: Modality = "native",
    ) -> None:
        json_schema = load_json_schema(json_schema)
        assert_valid_model_extraction(model)
        """
        Generate annotations from document files or in-memory documents
        and create a JSONL training set in one go.

        Args:
            json_schema: The JSON schema for validation
            documents: list of documents, each can be a Path/str or an IOBase object
            jsonl_path: Output path for the JSONL training file
            model: The model to use for processing
            temperature: Model temperature (0-1)
            batch_size: Number of examples to process in each batch
            max_concurrent: Maximum number of concurrent API calls
            root_dir: Where to store the per-document JSON annotations
        """

        async def process_example(doc: Path | str | IOBase, semaphore: asyncio.Semaphore) -> dict[str, Any]:
            """
            Process a single document (either a file path or an in-memory file-like object).
            Returns a dict with pointers to the original doc and the stored annotation JSON.
            """
            if isinstance(doc, (str, Path)):
                # Handle path or string
                doc_path = Path(doc)
                if not doc_path.is_file():
                    raise ValueError(f"Invalid file path: {doc_path}")

                # Extract results
                async with semaphore:
                    result = await self._client.documents.extractions.parse(
                        json_schema=json_schema,
                        document=doc_path,  # pass the actual Path to .extract
                        model=model,
                        temperature=temperature,
                        modality=modality,
                    )
                if result.choices[0].message.content is None:
                    print(f"Failed to extract content from {doc_path}")
                    return {"document_fpath": str(doc_path), "annotation_fpath": None}
                # Generate a unique filename for the annotation
                hash_str = hashlib.md5(doc_path.as_posix().encode()).hexdigest()
                annotation_path = Path(root_dir) / f"annotations_{hash_str}.json"

                annotation_path.parent.mkdir(parents=True, exist_ok=True)

                with open(annotation_path, 'w', encoding='utf-8') as f:
                    json.dump(result.choices[0].message.content, f, ensure_ascii=False, indent=2)

                return {"document_fpath": str(doc_path), "annotation_fpath": str(annotation_path)}

            elif isinstance(doc, IO):
                # Handle in-memory file-like object
                # 1) Read file content (but be careful with read pointer!)
                file_bytes = doc.read()

                # 2) Attempt to get a name; default to "uploaded_file" if none
                doc_name = getattr(doc, "name", "uploaded_file")

                # 3) Reset the file pointer if you plan to reuse `doc`
                #    (optional, depending on how you're using it)
                doc.seek(0)

                # 4) Call extract with the same doc object
                async with semaphore:
                    result = await self._client.documents.extractions.parse(
                        json_schema=json_schema,
                        document=doc,  # pass the IO object directly
                        model=model,
                        temperature=temperature,
                        modality=modality,
                    )

                if result.choices[0].message.content is None:
                    print(f"Failed to extract content from {doc_name}")
                    return {"document_fpath": doc_name, "annotation_fpath": None}

                # 5) Create a unique hash from the content
                hash_str = hashlib.md5(file_bytes).hexdigest()
                annotation_path = Path(root_dir) / f"annotations_{hash_str}.json"

                annotation_path.parent.mkdir(parents=True, exist_ok=True)

                with open(annotation_path, 'w', encoding='utf-8') as f:
                    json.dump(result.choices[0].message.content, f, ensure_ascii=False, indent=2)

                return {
                    "document_fpath": doc_name,  # or "in_memory_file"
                    "annotation_fpath": str(annotation_path),
                }

            else:
                raise ValueError(f"Unsupported document type: {type(doc)}")

        # Make sure output directory exists
        Path(root_dir).mkdir(parents=True, exist_ok=True)

        pairs_paths: list[dict[str, Path | str]] = []
        futures = []
        semaphore = asyncio.Semaphore(max_concurrent)
        # Split documents into batches
        for batch in tqdm([documents[i : i + batch_size] for i in range(0, len(documents), batch_size)], desc="Processing batches"):
            # Submit batch of tasks
            for doc in batch:
                futures.append(process_example(doc, semaphore))
            pairs_paths = await asyncio.gather(*futures)

        # Generate final training set from all results
        await self.save(json_schema=json_schema, pairs_paths=pairs_paths, jsonl_path=jsonl_path)

    async def benchmark(self, **kwargs: Any) -> None:
        """Benchmark model performance on a test dataset.

        Args:
            json_schema: JSON schema defining the expected data structure
            jsonl_path: Path to the JSONL file containing test examples
            model: The AI model to use for benchmarking
            temperature: Model temperature setting (0-1)

        Raises:
            NotImplementedError: This method is not implemented yet
        """

        # TODO

        raise NotImplementedError("Benchmarking is not implemented yet")

    async def filter(self, **kwargs: Any) -> None:
        """Filter examples from a JSONL file based on specified parameters.

        Args:
            json_schema: JSON schema defining the data structure
            jsonl_path: Path to the JSONL file to filter
            output_path: Optional path for the filtered output
            inplace: Whether to modify the file in place
            filter_parameters: Parameters to filter examples by (e.g., {"confidence": 0.8})

        Note:
            Filter parameters can include:
            - Number of tokens
            - Modality
            - Other custom parameters
        """
        raise NotImplementedError("Filtering is not implemented yet")

    async def print(self, jsonl_path: Path, output_path: Path = Path("annotations")) -> None:
        """Print a summary of the contents and statistics of a JSONL file.

        This method analyzes the JSONL file and displays various metrics and statistics
        about the dataset contents.

        Args:
            jsonl_path: Path to the JSONL file to analyze
            output_path: Directory where to save any generated reports
        """
        raise NotImplementedError("Printing is not implemented yet")

    async def stitch(self, **kwargs: Any) -> None:
        """Stitch annotations from a list of MIMEData objects into a single MIMEData object.

        This method combines multiple MIMEData annotations into a single object to avoid
        nested list structures (list[list[MIMEData]]) and maintain a simpler list[MIMEData] structure.

        Args:
            json_schema: The JSON schema for validation
            jsonl_path: Path to the JSONL file
            output_path: Optional path for the output file
            inplace: Whether to modify the file in place
            filter_parameters: Optional parameters for filtering
            modality: The modality to use for processing
        """

        raise NotImplementedError("Stitching is not implemented yet")



import base64
import io
import logging
from typing import List, Literal, Optional, Union, cast

import requests
from anthropic.types.content_block import ContentBlock
from anthropic.types.image_block_param import ImageBlockParam, Source
from anthropic.types.message_param import MessageParam
from anthropic.types.text_block_param import TextBlockParam
from anthropic.types.tool_result_block_param import ToolResultBlockParam
from anthropic.types.tool_use_block_param import ToolUseBlockParam
from google.genai.types import BlobDict, ContentDict, ContentUnionDict, PartDict  # type: ignore
from openai.types.chat.chat_completion_content_part_image_param import ChatCompletionContentPartImageParam
from openai.types.chat.chat_completion_content_part_input_audio_param import ChatCompletionContentPartInputAudioParam
from openai.types.chat.chat_completion_content_part_param import ChatCompletionContentPartParam
from openai.types.chat.chat_completion_content_part_text_param import ChatCompletionContentPartTextParam
from openai.types.chat.chat_completion_message_param import ChatCompletionMessageParam
from PIL import Image

from ..types.chat import ChatCompletionUiformMessage

MediaType = Literal["image/jpeg", "image/png", "image/gif", "image/webp"]


def convert_to_google_genai_format(messages: List[ChatCompletionUiformMessage]) -> tuple[str, list[ContentUnionDict]]:
    """
    Converts a list of ChatCompletionUiFormMessage to a format compatible with the google.genai SDK.


    Example:
        ```python
        import google.genai as genai

        # Configure the Gemini client
        genai.configure(api_key=os.environ["GEMINI_API_KEY"])

        # Initialize the model
        model = genai.GenerativeModel("gemini-2.0-flash")

        # Get messages in Gemini format
        gemini_messages = document_message.gemini_messages

        # Generate a response
        ```

    Args:
        messages (List[ChatCompletionUiformMessage]): List of chat messages.

    Returns:
        List[Union[Dict[str, str], str]]: A list of formatted inputs for the google.genai SDK.
    """
    system_message: str = ""
    formatted_content: list[ContentUnionDict] = []
    for message in messages:
        # -----------------------
        # Handle system message
        # -----------------------
        if message["role"] in ("system", "developer"):
            assert isinstance(message["content"], str), "System message content must be a string."
            if system_message != "":
                raise ValueError("Only one system message is allowed per chat.")
            system_message += message["content"]
            continue
        parts: list[PartDict] = []

        message_content = message['content']
        if isinstance(message_content, str):
            # Direct string content is treated as the prompt for the SDK
            parts.append(PartDict(text=message_content))
        elif isinstance(message_content, list):
            # Handle structured content
            for part in message_content:
                if part["type"] == "text":
                    parts.append(PartDict(text=part["text"]))
                elif part["type"] == "image_url":
                    url = part['image_url'].get('url', '')  # type: ignore
                    if url.startswith('data:image'):
                        # Extract base64 data and add it to the formatted inputs
                        media_type, data_content = url.split(";base64,")
                        media_type = media_type.split("data:")[-1]  # => "image/jpeg"
                        base64_data = data_content

                        # Try to convert to PIL.Image and append it to the formatted inputs
                        try:
                            image_bytes = base64.b64decode(base64_data)
                            parts.append(PartDict(inline_data=BlobDict(data=image_bytes, mime_type=media_type)))
                        except Exception:
                            pass
                elif part["type"] == "input_audio":
                    pass
                elif part["type"] == "file":
                    pass
                else:
                    pass

        formatted_content.append(ContentDict(parts=parts, role=("user" if message["role"] == "user" else "model")))

    return system_message, formatted_content


def convert_to_anthropic_format(messages: List[ChatCompletionUiformMessage]) -> tuple[str, List[MessageParam]]:
    """
    Converts a list of ChatCompletionUiformMessage to a format compatible with the Anthropic SDK.

    Args:
        messages (List[ChatCompletionUiformMessage]): List of chat messages.

    Returns:
        (system_message, formatted_messages):
            system_message (str | NotGiven):
                The system message if one was found, otherwise NOT_GIVEN.
            formatted_messages (List[MessageParam]):
                A list of formatted messages ready for Anthropic.
    """

    formatted_messages: list[MessageParam] = []
    system_message: str = ""

    for message in messages:
        content_blocks: list[Union[TextBlockParam, ImageBlockParam]] = []

        # -----------------------
        # Handle system message
        # -----------------------
        if message["role"] in ("system", "developer"):
            assert isinstance(message["content"], str), "System message content must be a string."
            if system_message != "":
                raise ValueError("Only one system message is allowed per chat.")
            system_message += message["content"]
            continue

        # -----------------------
        # Handle non-system roles
        # -----------------------
        if isinstance(message['content'], str):
            # Direct string content is treated as a single text block
            content_blocks.append(
                {
                    "type": "text",
                    "text": message['content'],
                }
            )

        elif isinstance(message['content'], list):
            # Handle structured content
            for part in message['content']:
                if part["type"] == "text":
                    part = cast(ChatCompletionContentPartTextParam, part)
                    content_blocks.append(
                        {
                            "type": "text",
                            "text": part['text'],  # type: ignore
                        }
                    )

                elif part["type"] == "input_audio":
                    part = cast(ChatCompletionContentPartInputAudioParam, part)
                    logging.warning("Audio input is not supported yet.")
                    # No blocks appended since not supported

                elif part["type"] == "image_url":
                    # Handle images that may be either base64 data-URLs or standard remote URLs
                    part = cast(ChatCompletionContentPartImageParam, part)
                    image_url = part["image_url"]["url"]

                    if "base64," in image_url:
                        # The string is already something like: data:image/jpeg;base64,xxxxxxxx...
                        media_type, data_content = image_url.split(";base64,")
                        # media_type might look like: "data:image/jpeg"
                        media_type = media_type.split("data:")[-1]  # => "image/jpeg"
                        base64_data = data_content
                    else:
                        # It's a remote URL, so fetch, encode, and derive media type from headers
                        try:
                            r = requests.get(image_url)
                            r.raise_for_status()
                            content_type = r.headers.get("Content-Type", "image/jpeg")
                            # fallback "image/jpeg" if no Content-Type given

                            # Only keep recognized image/* for anthropic
                            if content_type not in ("image/jpeg", "image/png", "image/gif", "image/webp"):
                                logging.warning(
                                    "Unrecognized Content-Type '%s' - defaulting to image/jpeg",
                                    content_type,
                                )
                                content_type = "image/jpeg"

                            media_type = content_type
                            base64_data = base64.b64encode(r.content).decode("utf-8")

                        except Exception:
                            logging.warning(
                                "Failed to load image from URL: %s",
                                image_url,
                                exc_info=True,
                                stack_info=True,
                            )
                            # Skip adding this block if error
                            continue

                    # Finally, append to content blocks
                    content_blocks.append(
                        {
                            "type": "image",
                            "source": {
                                "type": "base64",
                                "media_type": cast(MediaType, media_type),
                                "data": base64_data,
                            },
                        }
                    )

        formatted_messages.append(
            MessageParam(
                role=message["role"],  # type: ignore
                content=content_blocks,
            )
        )

    return system_message, formatted_messages


def convert_from_anthropic_format(messages: list[MessageParam], system_prompt: str) -> list[ChatCompletionUiformMessage]:
    """
    Converts a list of Anthropic MessageParam to a list of ChatCompletionUiformMessage.
    """
    formatted_messages: list[ChatCompletionUiformMessage] = [ChatCompletionUiformMessage(role="developer", content=system_prompt)]

    for message in messages:
        role = message["role"]
        content_blocks = message["content"]

        # Handle different content structures
        if isinstance(content_blocks, list) and len(content_blocks) == 1 and isinstance(content_blocks[0], dict) and content_blocks[0].get("type") == "text":
            # Simple text message
            formatted_messages.append(cast(ChatCompletionUiformMessage, {"role": role, "content": content_blocks[0].get("text", "")}))
        elif isinstance(content_blocks, list):
            # Message with multiple content parts or non-text content
            formatted_content: list[ChatCompletionContentPartParam] = []

            for block in content_blocks:
                if isinstance(block, dict):
                    if block.get("type") == "text":
                        formatted_content.append(cast(ChatCompletionContentPartParam, {"type": "text", "text": block.get("text", "")}))
                    elif block.get("type") == "image":
                        source = block.get("source", {})
                        if isinstance(source, dict) and source.get("type") == "base64":
                            # Convert base64 image to data URL format
                            media_type = source.get("media_type", "image/jpeg")
                            data = source.get("data", "")
                            image_url = f"data:{media_type};base64,{data}"

                            formatted_content.append(cast(ChatCompletionContentPartParam, {"type": "image_url", "image_url": {"url": image_url}}))

            formatted_messages.append(cast(ChatCompletionUiformMessage, {"role": role, "content": formatted_content}))

    return formatted_messages


def convert_to_openai_format(messages: List[ChatCompletionUiformMessage]) -> List[ChatCompletionMessageParam]:
    return cast(list[ChatCompletionMessageParam], messages)


def convert_from_openai_format(messages: list[ChatCompletionMessageParam]) -> list[ChatCompletionUiformMessage]:
    return cast(list[ChatCompletionUiformMessage], messages)


def separate_messages(
    messages: list[ChatCompletionUiformMessage],
) -> tuple[Optional[ChatCompletionUiformMessage], list[ChatCompletionUiformMessage], list[ChatCompletionUiformMessage]]:
    """
    Separates messages into system, user and assistant messages.

    Args:
        messages: List of chat messages containing system, user and assistant messages

    Returns:
        Tuple containing:
        - The system message if present, otherwise None
        - List of user messages
        - List of assistant messages
    """
    system_message = None
    user_messages = []
    assistant_messages = []

    for message in messages:
        if message["role"] in ("system", "developer"):
            system_message = message
        elif message["role"] == "user":
            user_messages.append(message)
        elif message["role"] == "assistant":
            assistant_messages.append(message)

    return system_message, user_messages, assistant_messages


def str_messages(messages: list[ChatCompletionUiformMessage], max_length: int = 100) -> str:
    """
    Converts a list of chat messages into a string representation with faithfully serialized structure.

    Args:
        messages (list[ChatCompletionUiformMessage]): The list of chat messages.
        max_length (int): Maximum length for content before truncation.

    Returns:
        str: A string representation of the messages with applied truncation.
    """

    def truncate(text: str, max_len: int) -> str:
        """Truncate text to max_len with ellipsis."""
        return text if len(text) <= max_len else f"{text[:max_len]}..."

    serialized: list[ChatCompletionUiformMessage] = []
    for message in messages:
        role = message["role"]
        content = message["content"]

        if isinstance(content, str):
            serialized.append({"role": role, "content": truncate(content, max_length)})
        elif isinstance(content, list):
            truncated_content: list[ChatCompletionContentPartParam] = []
            for part in content:
                if part["type"] == "text" and part["text"]:
                    truncated_content.append({"type": "text", "text": truncate(part["text"], max_length)})
                elif part["type"] == "image_url" and part["image_url"]:
                    image_url = part["image_url"].get("url", "unknown image")
                    truncated_content.append({"type": "image_url", "image_url": {"url": truncate(image_url, max_length)}})
            serialized.append({"role": role, "content": truncated_content})

    return repr(serialized)



from ..._resource import AsyncAPIResource, SyncAPIResource
from ...types.standards import PreparedRequest


class WebhookMixin:
    def prepare_create(self) -> PreparedRequest:
        return PreparedRequest(method="POST", url="/v1/secrets/webhook")

    def prepare_delete(self) -> PreparedRequest:
        return PreparedRequest(method="DELETE", url="/v1/secrets/webhook")


class Webhook(SyncAPIResource, WebhookMixin):
    """Webhook secret management wrapper"""

    def create(
        self,
    ) -> dict:
        """Create a webhook secret.

        Returns:
            dict: Response indicating success
        """
        request = self.prepare_create()
        response = self._client._prepared_request(request)

        return response

    def delete(self) -> dict:
        """Delete a webhook secret.

        Returns:
            dict: Response indicating success
        """
        request = self.prepare_delete()
        response = self._client._prepared_request(request)

        return response


class AsyncWebhook(AsyncAPIResource, WebhookMixin):
    """Webhook secret management wrapper"""

    async def create(self) -> dict:
        """Create a webhook secret.

        Returns:
            dict: Response indicating success
        """
        request = self.prepare_create()
        response = await self._client._prepared_request(request)
        return response

    async def delete(self) -> dict:
        """Delete a webhook secret.

        Returns:
            dict: Response indicating success
        """
        request = self.prepare_delete()
        response = await self._client._prepared_request(request)
        return response



import datetime
import json
from io import IOBase
from pathlib import Path
from typing import Any, Dict, Literal, Optional

import httpx
from openai.types.chat.chat_completion_reasoning_effort import ChatCompletionReasoningEffort
from PIL.Image import Image
from pydantic import HttpUrl

from ...._resource import AsyncAPIResource, SyncAPIResource
from ...._utils.ai_models import assert_valid_model_extraction
from ...._utils.mime import prepare_mime_document
from ....types.automations.links import Link, ListLinks, UpdateLinkRequest
from ....types.documents.extractions import UiParsedChatCompletion
from ....types.logs import AutomationLog, ExternalRequestLog, ListLogs
from ....types.mime import BaseMIMEData, MIMEData
from ....types.modalities import Modality
from ....types.standards import PreparedRequest


class LinksMixin:
    def prepare_create(
        self,
        name: str,
        json_schema: Dict[str, Any],
        webhook_url: HttpUrl,   
        webhook_headers: Optional[Dict[str, str]] = None,
        password: str | None = None,
        # DocumentExtraction Config
        image_resolution_dpi: Optional[int] = None,
        browser_canvas: Optional[str] = None,
        modality: Modality = "native",
        model: str = "gpt-4o-mini",
        temperature: float = 0,
        reasoning_effort: ChatCompletionReasoningEffort = "medium",
    ) -> PreparedRequest:
        assert_valid_model_extraction(model)

        data = {
            "name": name,
            "webhook_url": webhook_url,
            "webhook_headers": webhook_headers or {},
            "json_schema": json_schema,
            "password": password,
            "image_resolution_dpi": image_resolution_dpi,
            "browser_canvas": browser_canvas,
            "modality": modality,
            "model": model,
            "temperature": temperature,
            "reasoning_effort": reasoning_effort,
        }

        request = Link.model_validate(data)
        return PreparedRequest(method="POST", url="/v1/deployments/links", data=request.model_dump(mode='json'))

    def prepare_list(
        self,
        before: Optional[str] = None,
        after: Optional[str] = None,
        limit: Optional[int] = 10,
        order: Optional[Literal["asc", "desc"]] = "desc",
        # Filtering parameters
        link_id: Optional[str] = None,
        name: Optional[str] = None,
        webhook_url: Optional[str] = None,
        schema_id: Optional[str] = None,
        schema_data_id: Optional[str] = None,
    ) -> PreparedRequest:
        params = {
            "before": before,
            "after": after,
            "limit": limit,
            "order": order,
            "automation_id": link_id,
            "name": name,
            "webhook_url": webhook_url,
            "schema_id": schema_id,
            "schema_data_id": schema_data_id,
        }
        # Remove None values
        params = {k: v for k, v in params.items() if v is not None}

        return PreparedRequest(method="GET", url="/v1/deployments", params=params)

    def prepare_get(self, link_id: str) -> PreparedRequest:
        """Get a specific extraction link configuration.

        Args:
            link_id: ID of the extraction link

        Returns:
            Link: The extraction link configuration
        """
        return PreparedRequest(method="GET", url=f"/v1/deployments/{link_id}")

    def prepare_update(
        self,
        link_id: str,
        name: Optional[str] = None,
        webhook_url: Optional[HttpUrl] = None,
        webhook_headers: Optional[Dict[str, str]] = None,
        password: Optional[str] = None,
        image_resolution_dpi: Optional[int] = None,
        browser_canvas: Optional[str] = None,
        modality: Optional[Modality] = None,
        model: Optional[str] = None,
        temperature: Optional[float] = None,
        reasoning_effort: Optional[ChatCompletionReasoningEffort] = None,
        json_schema: Optional[Dict[str, Any]] = None,
    ) -> PreparedRequest:
        data: dict[str, Any] = {}

        if link_id is not None:
            data["id"] = link_id
        if name is not None:
            data["name"] = name
        if webhook_url is not None:
            data["webhook_url"] = webhook_url
        if webhook_headers is not None:
            data["webhook_headers"] = webhook_headers
        if password is not None:
            data["password"] = password
        if image_resolution_dpi is not None:
            data["image_resolution_dpi"] = image_resolution_dpi
        if browser_canvas is not None:
            data["browser_canvas"] = browser_canvas
        if modality is not None:
            data["modality"] = modality
        if model is not None:
            assert_valid_model_extraction(model)
            data["model"] = model
        if temperature is not None:
            data["temperature"] = temperature
        if json_schema is not None:
            data["json_schema"] = json_schema
        if reasoning_effort is not None:
            data["reasoning_effort"] = reasoning_effort

        request = UpdateLinkRequest.model_validate(data)
        return PreparedRequest(method="PUT", url=f"/v1/deployments/{link_id}", data=request.model_dump(mode='json'))

    def prepare_delete(self, link_id: str) -> PreparedRequest:
        return PreparedRequest(method="DELETE", url=f"/v1/processors/automations/links/{link_id}", raise_for_status=True)

class Links(SyncAPIResource, LinksMixin):
    """Extraction Link API wrapper for managing extraction link configurations"""

    def __init__(self, client: Any) -> None:
        super().__init__(client=client)

    def create(
        self,
        name: str,
        json_schema: Dict[str, Any],
        webhook_url: HttpUrl,
        webhook_headers: Optional[Dict[str, str]] = None,
        password: str | None = None,
        # DocumentExtraction Config
        image_resolution_dpi: Optional[int] = None,
        browser_canvas: Optional[str] = None,
        modality: Modality = "native",
        model: str = "gpt-4o-mini",
        temperature: float = 0,
        reasoning_effort: ChatCompletionReasoningEffort = "medium",
    ) -> Link:
        """Create a new extraction link configuration.

        Args:
            name: Name of the extraction link
            json_schema: JSON schema to validate extracted data
            webhook_url: Webhook endpoint for forwarding processed files
            webhook_headers: Optional HTTP headers for webhook requests
            password: Optional password for protected links
            image_resolution_dpi: Optional image resolution DPI
            browser_canvas: Optional browser canvas
            modality: Processing modality (currently only "native" supported)
            model: AI model to use for processing
            temperature: Model temperature setting
            reasoning_effort: The effort level for the model to reason about the input data.
        Returns:
            Link: The created extraction link configuration
        """

        request = self.prepare_create(name, json_schema, webhook_url, webhook_headers, password, image_resolution_dpi, browser_canvas, modality, model, temperature, reasoning_effort)
        response = self._client._prepared_request(request)

        print(f"Extraction Link Created. Link available at https://www.uiform.com/dashboard/processors/{response['id']}")
        return Link.model_validate(response)

    def list(
        self,
        before: Optional[str] = None,
        after: Optional[str] = None,
        limit: Optional[int] = 10,
        order: Optional[Literal["asc", "desc"]] = "desc",
        # Filtering parameters
        link_id: Optional[str] = None,
        name: Optional[str] = None,
        webhook_url: Optional[str] = None,
        schema_id: Optional[str] = None,
        schema_data_id: Optional[str] = None,
    ) -> ListLinks:
        """List extraction link configurations with pagination support.

        Args:
            before: Optional cursor for pagination before a specific link ID
            after: Optional cursor for pagination after a specific link ID
            limit: Optional limit on number of results (max 100)
            order: Optional sort order ("asc" or "desc")
            link_id: Optional filter by extraction link ID
            name: Optional filter by link name
            webhook_url: Optional filter by webhook URL
            schema_id: Optional filter by schema ID
            schema_data_id: Optional filter by schema data ID

        Returns:
            ListLinks: Paginated list of extraction link configurations with metadata
        """
        request = self.prepare_list(before, after, limit, order, link_id, name, webhook_url, schema_id, schema_data_id)
        response = self._client._prepared_request(request)
        return ListLinks.model_validate(response)

    def get(self, link_id: str) -> Link:
        """Get a specific extraction link configuration.

        Args:
            link_id: ID of the extraction link

        Returns:
            Link: The extraction link configuration
        """
        request = self.prepare_get(link_id)
        response = self._client._prepared_request(request)
        return Link.model_validate(response)

    def update(
        self,
        link_id: str,
        name: Optional[str] = None,
        webhook_url: Optional[HttpUrl] = None,
        webhook_headers: Optional[Dict[str, str]] = None,
        password: Optional[str] = None,
        image_resolution_dpi: Optional[int] = None,
        browser_canvas: Optional[str] = None,
        modality: Optional[Modality] = None,
        model: Optional[str] = None,
        temperature: Optional[float] = None,
        reasoning_effort: Optional[ChatCompletionReasoningEffort] = None,
        json_schema: Optional[Dict[str, Any]] = None,
    ) -> Link:
        """Update an extraction link configuration.

        Args:
            link_id: ID of the extraction link to update
            name: New name for the link
            webhook_url: New webhook endpoint URL
            webhook_headers: New webhook headers
            password: New password for protected links
            image_resolution_dpi: New image resolution DPI
            browser_canvas: New browser canvas
            modality: New processing modality
            model: New AI model
            temperature: New temperature setting
            reasoning_effort: The effort level for the model to reason about the input data.
            json_schema: New JSON schema

        Returns:
            Link: The updated extraction link configuration
        """

        request = self.prepare_update(link_id, name, webhook_url, webhook_headers, password, image_resolution_dpi, browser_canvas, modality, model, temperature, reasoning_effort, json_schema)
        response = self._client._prepared_request(request)
        return Link.model_validate(response)

    def delete(self, link_id: str) -> None:
        """Delete an extraction link configuration.

        Args:
            link_id: ID of the extraction link to delete

        Returns:
            Dict[str, str]: Response message confirming deletion
        """
        request = self.prepare_delete(link_id)
        self._client._prepared_request(request)

class AsyncLinks(AsyncAPIResource, LinksMixin):
    """Async Extraction Link API wrapper for managing extraction link configurations"""

    def __init__(self, client: Any) -> None:
        super().__init__(client=client)

    async def create(
        self,
        name: str,
        json_schema: Dict[str, Any],
        webhook_url: HttpUrl,
        webhook_headers: Optional[Dict[str, str]] = None,
        password: str | None = None,
        image_resolution_dpi: Optional[int] = None,
        browser_canvas: Optional[str] = None,
        modality: Modality = "native",
        model: str = "gpt-4o-mini",
        temperature: float = 0,
        reasoning_effort: ChatCompletionReasoningEffort = "medium",
    ) -> Link:
        request = self.prepare_create(name, json_schema, webhook_url, webhook_headers, password, image_resolution_dpi, browser_canvas, modality, model, temperature, reasoning_effort)
        response = await self._client._prepared_request(request)
        print(f"Extraction Link Created. Link available at https://www.uiform.com/dashboard/processors/{response['id']}")
        return Link.model_validate(response)

    async def list(
        self,
        before: Optional[str] = None,
        after: Optional[str] = None,
        limit: Optional[int] = 10,
        order: Optional[Literal["asc", "desc"]] = "desc",
        link_id: Optional[str] = None,
        name: Optional[str] = None,
        webhook_url: Optional[str] = None,
        schema_id: Optional[str] = None,
        schema_data_id: Optional[str] = None,
    ) -> ListLinks:
        request = self.prepare_list(before, after, limit, order, link_id, name, webhook_url, schema_id, schema_data_id)
        response = await self._client._prepared_request(request)
        return ListLinks.model_validate(response)

    async def get(self, link_id: str) -> Link:
        request = self.prepare_get(link_id)
        response = await self._client._prepared_request(request)
        return Link.model_validate(response)

    async def update(
        self,
        link_id: str,
        name: Optional[str] = None,
        webhook_url: Optional[HttpUrl] = None,
        webhook_headers: Optional[Dict[str, str]] = None,
        password: Optional[str] = None,
        image_resolution_dpi: Optional[int] = None,
        browser_canvas: Optional[str] = None,
        modality: Optional[Modality] = None,
        model: Optional[str] = None,
        temperature: Optional[float] = None,
        reasoning_effort: Optional[ChatCompletionReasoningEffort] = None,
        json_schema: Optional[Dict[str, Any]] = None,
    ) -> Link:
        request = self.prepare_update(link_id, name, webhook_url, webhook_headers, password, image_resolution_dpi, browser_canvas, modality, model, temperature, reasoning_effort, json_schema)
        response = await self._client._prepared_request(request)
        return Link.model_validate(response)

    async def delete(self, link_id: str) -> None:
        request = self.prepare_delete(link_id)
        await self._client._prepared_request(request)



from typing import Iterable, Literal, TypedDict, Union

from openai.types.chat.chat_completion_content_part_param import ChatCompletionContentPartParam


class ChatCompletionUiformMessage(TypedDict):  # homemade replacement for ChatCompletionMessageParam because iterable messes the serialization with pydantic
    role: Literal['user', 'system', 'assistant', 'developer']
    content: Union[str, list[ChatCompletionContentPartParam]]
